# ActivitySim
# See full license in LICENSE.txt.
import os.path
import logging
import pytest

import pandas as pd

from .. import tracing
from .. import inject
from .. import orca


def close_handlers():

    loggers = logging.Logger.manager.loggerDict
    for name in loggers:
        logger = logging.getLogger(name)
        logger.handlers = []
        logger.propagate = True
        logger.setLevel(logging.NOTSET)


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def add_canonical_dirs():

    inject.clear_cache()

    configs_dir = os.path.join(os.path.dirname(__file__), 'configs')
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), 'output')
    inject.add_injectable("output_dir", output_dir)


def test_config_logger(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    logger = logging.getLogger('activitysim')

    file_handlers = [h for h in logger.handlers if type(h) is logging.FileHandler]
    assert len(file_handlers) == 1
    asim_logger_baseFilename = file_handlers[0].baseFilename

    print("handlers:", logger.handlers)

    logger.info('test_config_logger')
    logger.info('log_info')
    logger.warning('log_warn1')

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert "could not find conf file" not in out
    assert 'log_warn1' in out
    assert 'log_info' not in out

    close_handlers()

    logger = logging.getLogger(__name__)
    logger.warning('log_warn2')

    with open(asim_logger_baseFilename, 'r') as content_file:
        content = content_file.read()
        print(content)
    assert 'log_warn1' in content
    assert 'log_warn2' not in content


def test_print_summary(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    tracing.print_summary('label', df=pd.DataFrame(), describe=False, value_counts=False)

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert 'print_summary neither value_counts nor describe' in out

    close_handlers()


def test_register_households(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    df = pd.DataFrame({'zort': ['a', 'b', 'c']}, index=[1, 2, 3])

    inject.add_injectable('traceable_tables', ['households'])
    inject.add_injectable("trace_hh_id", 5)

    tracing.register_traceable_table('households', df)
    out, err = capsys.readouterr()
    # print out   # don't consume output

    assert "Can't register table 'households' without index name" in out

    df.index.name = 'household_id'
    tracing.register_traceable_table('households', df)
    out, err = capsys.readouterr()
    # print out   # don't consume output

    # should warn that household id not in index
    assert 'trace_hh_id 5 not in dataframe' in out

    close_handlers()


def test_register_tours(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    inject.add_injectable('traceable_tables', ['households', 'tours'])

    # in case another test injected this
    inject.add_injectable("trace_tours", [])
    inject.add_injectable("trace_hh_id", 3)  # need this or register_traceable_table is a nop

    tours_df = pd.DataFrame({'zort': ['a', 'b', 'c']}, index=[10, 11, 12])
    tours_df.index.name = 'tour_id'

    tracing.register_traceable_table('tours', tours_df)

    out, err = capsys.readouterr()
    assert "can't find a registered table to slice table 'tours' index name 'tour_id'" in out

    inject.add_injectable("trace_hh_id", 3)
    households_df = pd.DataFrame({'dzing': ['a', 'b', 'c']}, index=[1, 2, 3])
    households_df.index.name = 'household_id'
    tracing.register_traceable_table('households', households_df)

    tracing.register_traceable_table('tours', tours_df)

    out, err = capsys.readouterr()
    # print out  # don't consume output
    assert "can't find a registered table to slice table 'tours'" in out

    tours_df['household_id'] = [1, 5, 3]

    tracing.register_traceable_table('tours', tours_df)

    out, err = capsys.readouterr()
    print(out)  # don't consume output

    # should be tracing tour with tour_id 3
    traceable_table_ids = inject.get_injectable('traceable_table_ids')
    assert traceable_table_ids['tours'] == [12]

    close_handlers()


def test_write_csv(capsys):

    add_canonical_dirs()

    tracing.config_logger()

    # should complain if df not a DataFrame or Series
    tracing.write_csv(df='not a df or series', file_name='baddie')

    out, err = capsys.readouterr()

    print(out)  # don't consume output

    assert "unexpected type" in out

    close_handlers()


def test_slice_ids():

    df = pd.DataFrame({'household_id': [1, 2, 3]}, index=[11, 12, 13])

    # slice by named column
    sliced_df = tracing.slice_ids(df, [1, 3, 6], column='household_id')
    assert len(sliced_df.index) == 2

    # slice by index
    sliced_df = tracing.slice_ids(df, [6, 12], column=None)
    assert len(sliced_df.index) == 1

    # attempt to slice by non-existent column
    with pytest.raises(RuntimeError) as excinfo:
        sliced_df = tracing.slice_ids(df, [5, 6], column='baddie')
    assert "slice_ids slicer column 'baddie' not in dataframe" in str(excinfo.value)


def test_basic(capsys):

    close_handlers()

    configs_dir = os.path.join(os.path.dirname(__file__), 'configs')
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), 'output')
    inject.add_injectable("output_dir", output_dir)

    # remove existing handlers or basicConfig is a NOP
    logging.getLogger().handlers = []

    tracing.config_logger(basic=True)

    logger = logging.getLogger()
    file_handlers = [h for h in logger.handlers if type(h) is logging.FileHandler]
    assert len(file_handlers) == 0

    logger = logging.getLogger('activitysim')

    logger.info('test_basic')
    logger.debug('log_debug')
    logger.info('log_info')
    logger.warning('log_warn')

    out, err = capsys.readouterr()

    # don't consume output
    print(out)

    assert 'log_warn' in out
    assert 'log_info' in out
    assert 'log_debug' not in out

    close_handlers()



# ActivitySim
# See full license in LICENSE.txt.

import logging
import warnings
import os

import pandas as pd

from activitysim.core import (
    inject,
    config,
    util,
)
from activitysim.core import mem

logger = logging.getLogger(__name__)


def read_input_table(tablename):
    """Reads input table name and returns cleaned DataFrame.

    Uses settings found in input_table_list in global settings file

    Parameters
    ----------
    tablename : string

    Returns
    -------
    pandas DataFrame
    """
    table_list = config.setting('input_table_list')
    assert table_list is not None, 'no input_table_list found in settings'

    table_info = None
    for info in table_list:
        if info['tablename'] == tablename:
            table_info = info

    assert table_info is not None, \
        f"could not find info for for tablename {tablename} in settings file"

    return read_from_table_info(table_info)


def read_from_table_info(table_info):
    """
    Read input text files and return cleaned up DataFrame.

    table_info is a dictionary that specifies the following input params.

    See input_table_list in settings.yaml in the example folder for a working example

    +--------------+----------------------------------------------------------+
    | key          | description                                              |
    +==============+=========================================+================+
    | tablename    | name of pipeline table in which to store dataframe       |
    +--------------+----------------------------------------------------------+
    | filename     | name of csv file to read (in data_dir)                   |
    +--------------+----------------------------------------------------------+
    | column_map   | list of input columns to rename from_name: to_name       |
    +--------------+----------------------------------------------------------+
    | index_col    | name of column to set as dataframe index column          |
    +--------------+----------------------------------------------------------+
    | drop_columns | list of column names of columns to drop                  |
    +--------------+----------------------------------------------------------+
    | h5_tablename | name of target table in HDF5 file                        |
    +--------------+----------------------------------------------------------+

    """
    input_store = config.setting('input_store', None)
    create_input_store = config.setting('create_input_store', default=False)

    tablename = table_info.get('tablename')
    data_filename = table_info.get('filename', input_store)
    h5_tablename = table_info.get('h5_tablename') or tablename
    drop_columns = table_info.get('drop_columns', None)
    column_map = table_info.get('column_map', None)
    keep_columns = table_info.get('keep_columns', None)
    rename_columns = table_info.get('rename_columns', None)
    index_col = table_info.get('index_col', None)

    assert tablename is not None, 'no tablename provided'
    assert data_filename is not None, 'no input file provided'

    data_file_path = config.data_file_path(data_filename)

    df = _read_input_file(data_file_path, h5_tablename=h5_tablename)

    # logger.debug('raw %s table columns: %s' % (tablename, df.columns.values))
    logger.debug('raw %s table size: %s' % (tablename, util.df_size(df)))

    if create_input_store:
        h5_filepath = config.output_file_path('input_data.h5')
        logger.info('writing %s to %s' % (h5_tablename, h5_filepath))
        df.to_hdf(h5_filepath, key=h5_tablename, mode='a')

        csv_dir = config.output_file_path('input_data')
        if not os.path.exists(csv_dir):
            os.makedirs(csv_dir)  # make directory if needed
        df.to_csv(os.path.join(csv_dir, '%s.csv' % tablename), index=False)

    if drop_columns:
        logger.debug("dropping columns: %s" % drop_columns)
        df.drop(columns=drop_columns, inplace=True, errors='ignore')

    if column_map:
        warnings.warn("table_inf option 'column_map' renamed 'rename_columns'"
                      "Support for 'column_map' will be removed in future versions.",
                      FutureWarning)
        logger.debug("renaming columns: %s" % column_map)
        df.rename(columns=column_map, inplace=True)

    # rename columns first, so keep_columns can be a stable list of expected/required columns
    if rename_columns:
        logger.debug("renaming columns: %s" % rename_columns)
        df.rename(columns=rename_columns, inplace=True)

    # set index
    if index_col is not None:
        if index_col in df.columns:
            assert not df.duplicated(index_col).any()
            df.set_index(index_col, inplace=True)
        else:
            # FIXME not sure we want to do this. More likely they omitted index col than that they want to name it?
            # df.index.names = [index_col]
            logger.error(f"index_col '{index_col}' specified in configs but not in {tablename} table!")
            logger.error(f"{tablename} columns are: {list(df.columns)}")
            raise RuntimeError(f"index_col '{index_col}' not in {tablename} table!")

    if keep_columns:
        logger.debug("keeping columns: %s" % keep_columns)
        if not set(keep_columns).issubset(set(df.columns)):
            logger.error(f"Required columns missing from {tablename} table: "
                         f"{list(set(keep_columns).difference(set(df.columns)))}")
            logger.error(f"{tablename} table has columns: {list(df.columns)}")
            raise RuntimeError(f"Required columns missing from {tablename} table")

        df = df[keep_columns]

    if df.columns.duplicated().any():
        duplicate_column_names = df.columns[df.columns.duplicated(keep=False)].unique().to_list()
        assert not df.columns.duplicated().any(), f"duplicate columns names in {tablename}: {duplicate_column_names}"

    logger.debug('%s table columns: %s' % (tablename, df.columns.values))
    logger.debug('%s table size: %s' % (tablename, util.df_size(df)))
    logger.info('%s index name: %s' % (tablename, df.index.name))

    return df


def _read_input_file(filepath, h5_tablename=None):
    assert os.path.exists(filepath), 'input file not found: %s' % filepath

    if filepath.endswith('.csv'):
        return _read_csv_with_fallback_encoding(filepath)

    if filepath.endswith('.h5'):
        assert h5_tablename is not None, 'must provide a tablename to read HDF5 table'
        logger.info('reading %s table from %s' % (h5_tablename, filepath))
        return pd.read_hdf(filepath, h5_tablename)

    raise IOError(
        'Unsupported file type: %s. '
        'ActivitySim supports CSV and HDF5 files only' % filepath)


def _read_csv_with_fallback_encoding(filepath):
    """read a CSV to a pandas DataFrame using default utf-8 encoding,
    but try alternate Windows-compatible cp1252 if unicode fails

    """

    try:
        logger.info('Reading CSV file %s' % filepath)
        return pd.read_csv(filepath, comment='#')
    except UnicodeDecodeError:
        logger.warning(
            'Reading %s with default utf-8 encoding failed, trying cp1252 instead', filepath)
        return pd.read_csv(filepath, comment='#', encoding='cp1252')



# ActivitySim
# See full license in LICENSE.txt.
from builtins import range

import logging

import numpy as np
import pandas as pd

from activitysim.core import logit
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import tracing
from activitysim.core import chunk
from activitysim.core import pipeline

from activitysim.core.util import reindex

from activitysim.abm.models.util.trip import failed_trip_cohorts
from activitysim.abm.models.util.trip import cleanup_failed_trips


logger = logging.getLogger(__name__)

"""
StopDepartArrivePeriodModel

StopDepartArriveProportions.csv
tourpurp,isInbound,interval,trip,p1,p2,p3,p4,p5...p40

"""

NO_TRIP_ID = 0
NO_DEPART = 0

DEPART_ALT_BASE = 'DEPART_ALT_BASE'

FAILFIX = 'FAILFIX'
FAILFIX_CHOOSE_MOST_INITIAL = 'choose_most_initial'
FAILFIX_DROP_AND_CLEANUP = 'drop_and_cleanup'
FAILFIX_DEFAULT = FAILFIX_CHOOSE_MOST_INITIAL

PROBS_JOIN_COLUMNS = ['primary_purpose', 'outbound', 'tour_hour', 'trip_num']


def set_tour_hour(trips, tours):
    """
    add columns 'tour_hour', 'earliest', 'latest' to trips

    Parameters
    ----------
    trips: pd.DataFrame
    tours: pd.DataFrame

    Returns
    -------
    modifies trips in place
    """

    # all trips must depart between tour start and end
    trips['earliest'] = reindex(tours.start, trips.tour_id)
    trips['latest'] = reindex(tours.end, trips.tour_id)

    # tour_hour is start for outbound trips, and end for inbound trips
    trips['tour_hour'] = np.where(
        trips.outbound,
        trips['earliest'],
        trips['latest']).astype(np.int8)

    # subtours indexed by parent_tour_id
    subtours = tours.loc[tours.primary_purpose == 'atwork',
                         ['tour_num', 'tour_count', 'parent_tour_id', 'start', 'end']]

    subtours.parent_tour_id = subtours.parent_tour_id.astype(np.int64)
    subtours = subtours.set_index('parent_tour_id')
    subtours = subtours.astype(np.int16)  # remaining columns are all small ints

    # bool series
    trip_has_subtours = trips.tour_id.isin(subtours.index)

    outbound = trip_has_subtours & trips.outbound
    trips.loc[outbound, 'latest'] = \
        reindex(subtours[subtours.tour_num == 1]['start'], trips[outbound].tour_id)

    inbound = trip_has_subtours & ~trips.outbound
    trips.loc[inbound, 'earliest'] = \
        reindex(subtours[subtours.tour_num == subtours.tour_count]['end'], trips[inbound].tour_id)


def clip_probs(trips, probs, model_settings):
    """
    zero out probs before trips.earliest or after trips.latest

    Parameters
    ----------
    trips: pd.DataFrame
    probs: pd.DataFrame
        one row per trip, one column per time period, with float prob of picking that time period

    depart_alt_base: int
        int to add to probs column index to get time period it represents.
        e.g. depart_alt_base = 5 means first column (column 0) represents 5 am

    Returns
    -------
    probs: pd.DataFrame
        clipped version of probs

    """

    depart_alt_base = model_settings.get(DEPART_ALT_BASE)

    # there should be one row in probs per trip
    assert trips.shape[0] == probs.shape[0]

    # probs should sum to 1 across rows before clipping
    probs = probs.div(probs.sum(axis=1), axis=0)

    num_rows, num_cols = probs.shape
    ix_map = np.tile(np.arange(0, num_cols), num_rows).reshape(num_rows, num_cols) + depart_alt_base
    # 5 6 7 8 9 10...
    # 5 6 7 8 9 10...
    # 5 6 7 8 9 10...

    clip_mask = ((ix_map >= trips.earliest.values.reshape(num_rows, 1)) &
                 (ix_map <= trips.latest.values.reshape(num_rows, 1))) * 1
    #  [0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0]
    #  [0 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0]
    #  [0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0]...

    probs = probs*clip_mask

    return probs


def report_bad_choices(bad_row_map, df, filename, trace_label, trace_choosers=None):
    """

    Parameters
    ----------
    bad_row_map
    df : pandas.DataFrame
        utils or probs dataframe
    trace_choosers : pandas.dataframe
        the choosers df (for interaction_simulate) to facilitate the reporting of hh_id
        because we  can't deduce hh_id from the interaction_dataset which is indexed on index
        values from alternatives df

    """

    df = df[bad_row_map]
    if trace_choosers is None:
        hh_ids = tracing.hh_id_for_chooser(df.index, df)
    else:
        hh_ids = tracing.hh_id_for_chooser(df.index, trace_choosers)
    df['household_id'] = hh_ids

    filename = "%s.%s" % (trace_label, filename)

    logger.info("dumping %s" % filename)
    tracing.write_csv(df, file_name=filename, transpose=False)

    # log the indexes of the first MAX_PRINT offending rows
    MAX_PRINT = 0
    for idx in df.index[:MAX_PRINT].values:

        row_msg = "%s : failed %s = %s (hh_id = %s)" % \
                  (trace_label, df.index.name, idx, df.household_id.loc[idx])

        logger.warning(row_msg)


def schedule_nth_trips(
        trips,
        probs_spec,
        model_settings,
        first_trip_in_leg,
        report_failed_trips,
        trace_hh_id,
        trace_label):
    """
    We join each trip with the appropriate row in probs_spec by joining on probs_join_cols,
    which should exist in both trips, probs_spec dataframe.

    Parameters
    ----------
    trips: pd.DataFrame
    probs_spec: pd.DataFrame
        Dataframe of probs for choice of depart times and join columns to match them with trips.
        Depart columns names are irrelevant. Instead, they are position dependent,
        time period choice is their index + depart_alt_base
    depart_alt_base: int
        int to add to probs column index to get time period it represents.
        e.g. depart_alt_base = 5 means first column (column 0) represents 5 am
    report_failed_trips : bool
    trace_hh_id
    trace_label

    Returns
    -------
    choices: pd.Series
        time periods depart choices, one per trip (except for trips with zero probs)
    """

    depart_alt_base = model_settings.get('DEPART_ALT_BASE')

    probs_cols = [c for c in probs_spec.columns if c not in PROBS_JOIN_COLUMNS]

    # left join trips to probs (there may be multiple rows per trip for multiple depart ranges)
    choosers = pd.merge(trips.reset_index(), probs_spec, on=PROBS_JOIN_COLUMNS,
                        how='left').set_index('trip_id')
    chunk.log_df(trace_label, "choosers", choosers)

    if trace_hh_id and tracing.has_trace_targets(trips):
        tracing.trace_df(choosers, '%s.choosers' % trace_label)

    # choosers should now match trips row for row
    assert choosers.index.is_unique
    assert len(choosers.index) == len(trips.index)

    # zero out probs outside earliest-latest window
    chooser_probs = clip_probs(trips, choosers[probs_cols], model_settings)
    chunk.log_df(trace_label, "chooser_probs", chooser_probs)

    if first_trip_in_leg:
        # probs should sum to 1 unless all zero
        chooser_probs = chooser_probs.div(chooser_probs.sum(axis=1), axis=0).fillna(0)

    # probs should sum to 1 with residual probs resulting in choice of 'fail'
    chooser_probs['fail'] = 1 - chooser_probs.sum(axis=1).clip(0, 1)
    chunk.log_df(trace_label, "chooser_probs", chooser_probs)

    if trace_hh_id and tracing.has_trace_targets(trips):
        tracing.trace_df(chooser_probs, '%s.chooser_probs' % trace_label)

    choices, rands = logit.make_choices(chooser_probs, trace_label=trace_label, trace_choosers=choosers)

    chunk.log_df(trace_label, "choices", choices)
    chunk.log_df(trace_label, "rands", rands)

    if trace_hh_id and tracing.has_trace_targets(trips):
        tracing.trace_df(choices, '%s.choices' % trace_label, columns=[None, 'depart'])
        tracing.trace_df(rands, '%s.rands' % trace_label, columns=[None, 'rand'])

    # convert alt choice index to depart time (setting failed choices to -1)
    failed = (choices == chooser_probs.columns.get_loc('fail'))
    choices = (choices + depart_alt_base).where(~failed, -1)

    chunk.log_df(trace_label, "failed", failed)

    # report failed trips while we have the best diagnostic info
    if report_failed_trips and failed.any():
        report_bad_choices(
            bad_row_map=failed,
            df=choosers,
            filename='failed_choosers',
            trace_label=trace_label,
            trace_choosers=None)

    # trace before removing failures
    if trace_hh_id and tracing.has_trace_targets(trips):
        tracing.trace_df(choices, '%s.choices' % trace_label, columns=[None, 'depart'])
        tracing.trace_df(rands, '%s.rands' % trace_label, columns=[None, 'rand'])

    # remove any failed choices
    if failed.any():
        choices = choices[~failed]

    assert (choices >= trips.earliest[~failed]).all()
    assert (choices <= trips.latest[~failed]).all()

    return choices


def schedule_trips_in_leg(
        outbound,
        trips,
        probs_spec,
        model_settings,
        last_iteration,
        trace_hh_id, trace_label):
    """

    Parameters
    ----------
    outbound
    trips
    probs_spec
    depart_alt_base
    last_iteration
    trace_hh_id
    trace_label

    Returns
    -------
    choices: pd.Series
        depart choice for trips, indexed by trip_id
    """

    failfix = model_settings.get(FAILFIX, FAILFIX_DEFAULT)

    # logger.debug("%s scheduling %s trips" % (trace_label, trips.shape[0]))

    assert len(trips) > 0

    assert (trips.outbound == outbound).all()

    # initial trip of leg and all atwork trips get tour_hour
    is_initial = (trips.trip_num == 1) if outbound else (trips.trip_num == trips.trip_count)
    no_scheduling = is_initial | (trips.primary_purpose == 'atwork')
    choices = trips.tour_hour[no_scheduling]

    if no_scheduling.all():
        return choices

    result_list = []
    result_list.append(choices)
    trips = trips[~no_scheduling]

    # add next_trip_id temp column (temp as trips is now a copy, as result of slicing)
    trips = trips.sort_index()
    trips['next_trip_id'] = np.roll(trips.index, -1 if outbound else 1)
    is_final = (trips.trip_num == trips.trip_count) if outbound else (trips.trip_num == 1)
    trips.next_trip_id = trips.next_trip_id.where(~is_final, NO_TRIP_ID)

    # iterate over outbound trips in ascending trip_num order, skipping the initial trip
    # iterate over inbound trips in descending trip_num order, skipping the finial trip
    first_trip_in_leg = True
    for i in range(trips.trip_num.min(), trips.trip_num.max() + 1):

        if outbound:
            nth_trips = trips[trips.trip_num == i]
        else:
            nth_trips = trips[trips.trip_num == trips.trip_count - i]

        nth_trace_label = tracing.extend_trace_label(trace_label, 'num_%s' % i)

        with chunk.chunk_log(nth_trace_label):
            choices = schedule_nth_trips(
                nth_trips,
                probs_spec,
                model_settings,
                first_trip_in_leg=first_trip_in_leg,
                report_failed_trips=last_iteration,
                trace_hh_id=trace_hh_id,
                trace_label=nth_trace_label)

        # if outbound, this trip's depart constrains next trip's earliest depart option
        # if inbound, we are handling in reverse order, so it constrains latest depart instead
        ADJUST_NEXT_DEPART_COL = 'earliest' if outbound else 'latest'

        # most initial departure (when no choice was made because all probs were zero)
        if last_iteration and (failfix == FAILFIX_CHOOSE_MOST_INITIAL):
            choices = choices.reindex(nth_trips.index)
            logger.warning("%s coercing %s depart choices to most initial" %
                           (nth_trace_label, choices.isna().sum()))
            choices = choices.fillna(trips[ADJUST_NEXT_DEPART_COL])

        # adjust allowed depart range of next trip
        has_next_trip = (nth_trips.next_trip_id != NO_TRIP_ID)
        if has_next_trip.any():
            next_trip_ids = nth_trips.next_trip_id[has_next_trip]
            # patch choice any trips with next_trips that weren't scheduled
            trips.loc[next_trip_ids, ADJUST_NEXT_DEPART_COL] = \
                choices.reindex(next_trip_ids.index).fillna(trips[ADJUST_NEXT_DEPART_COL]).values

        result_list.append(choices)

        first_trip_in_leg = False

    if len(result_list) > 1:
        choices = pd.concat(result_list)

    return choices


def trip_scheduling_calc_row_size(trips, spec, trace_label):

    sizer = chunk.RowSizeEstimator(trace_label)

    # NOTE we chunk chunk_id
    # scale row_size by average number of chooser rows per chunk_id
    num_choosers = trips['chunk_id'].max() + 1
    rows_per_chunk_id = len(trips) / num_choosers

    # only non-initial trips require scheduling, segment handing first such trip in tour will use most space
    outbound_chooser = (trips.trip_num == 2) & trips.outbound & (trips.primary_purpose != 'atwork')
    inbound_chooser = (trips.trip_num == trips.trip_count-1) & ~trips.outbound & (trips.primary_purpose != 'atwork')

    # furthermore, inbound and outbound are scheduled independently
    if outbound_chooser.sum() > inbound_chooser.sum():
        is_chooser = outbound_chooser
        logger.debug(f"{trace_label} {is_chooser.sum()} outbound_choosers of {len(trips)} require scheduling")
    else:
        is_chooser = inbound_chooser
        logger.debug(f"{trace_label} {is_chooser.sum()} inbound_choosers of {len(trips)} require scheduling")

    chooser_fraction = is_chooser.sum()/len(trips)
    logger.debug(f"{trace_label} chooser_fraction {chooser_fraction *100}%")

    chooser_row_size = len(trips.columns) + len(spec.columns) - len(PROBS_JOIN_COLUMNS)
    sizer.add_elements(chooser_fraction * chooser_row_size, 'choosers')

    # might be clipped to fewer but this is worst case
    chooser_probs_row_size = len(spec.columns) - len(PROBS_JOIN_COLUMNS)
    sizer.add_elements(chooser_fraction * chooser_probs_row_size, 'chooser_probs')

    sizer.add_elements(chooser_fraction, 'choices')
    sizer.add_elements(chooser_fraction, 'rands')
    sizer.add_elements(chooser_fraction, 'failed')

    row_size = sizer.get_hwm()
    row_size = row_size * rows_per_chunk_id

    return row_size


def run_trip_scheduling(
        trips,
        tours,
        probs_spec,
        model_settings,
        last_iteration,
        chunk_size,
        trace_hh_id,
        trace_label):

    set_tour_hour(trips, tours)

    row_size = chunk_size and trip_scheduling_calc_row_size(trips, probs_spec, trace_label)

    # only non-initial trips require scheduling, segment handing first such trip in tour will use most space
    # is_outbound_chooser = (trips.trip_num > 1) & trips.outbound & (trips.primary_purpose != 'atwork')
    # is_inbound_chooser = (trips.trip_num < trips.trip_count) & ~trips.outbound & (trips.primary_purpose != 'atwork')
    # num_choosers = (is_inbound_chooser | is_outbound_chooser).sum()

    result_list = []
    for i, trips_chunk, chunk_trace_label \
            in chunk.adaptive_chunked_choosers_by_chunk_id(trips, chunk_size, row_size, trace_label):

        if trips_chunk.outbound.any():
            leg_trace_label = tracing.extend_trace_label(chunk_trace_label, 'outbound')
            with chunk.chunk_log(leg_trace_label):
                choices = \
                    schedule_trips_in_leg(
                        outbound=True,
                        trips=trips_chunk[trips_chunk.outbound],
                        probs_spec=probs_spec,
                        model_settings=model_settings,
                        last_iteration=last_iteration,
                        trace_hh_id=trace_hh_id,
                        trace_label=leg_trace_label)
                result_list.append(choices)

        if (~trips_chunk.outbound).any():
            leg_trace_label = tracing.extend_trace_label(chunk_trace_label, 'inbound')
            with chunk.chunk_log(leg_trace_label):
                choices = \
                    schedule_trips_in_leg(
                        outbound=False,
                        trips=trips_chunk[~trips_chunk.outbound],
                        probs_spec=probs_spec,
                        model_settings=model_settings,
                        last_iteration=last_iteration,
                        trace_hh_id=trace_hh_id,
                        trace_label=leg_trace_label)
                result_list.append(choices)

    choices = pd.concat(result_list)

    return choices


@inject.step()
def trip_scheduling(
        trips,
        tours,
        chunk_size,
        trace_hh_id):

    """
    Trip scheduling assigns depart times for trips within the start, end limits of the tour.

    The algorithm is simplistic:

    The first outbound trip starts at the tour start time, and subsequent outbound trips are
    processed in trip_num order, to ensure that subsequent trips do not depart before the
    trip that preceeds them.

    Inbound trips are handled similarly, except in reverse order, starting with the last trip,
    and working backwards to ensure that inbound trips do not depart after the trip that
    succeeds them.

    The probability spec assigns probabilities for depart times, but those possible departs must
    be clipped to disallow depart times outside the tour limits, the departs of prior trips, and
    in the case of work tours, the start/end times of any atwork subtours.

    Scheduling can fail if the probability table assigns zero probabilities to all the available
    depart times in a trip's depart window. (This could be avoided by giving every window a small
    probability, rather than zero, but the existing mtctm1 prob spec does not do this. I believe
    this is due to the its having been generated from a small household travel survey sample
    that lacked any departs for some time periods.)

    Rescheduling the trips that fail (along with their inbound or outbound leg-mates) can sometimes
    fix this problem, if it was caused by an earlier trip's depart choice blocking a subsequent
    trip's ability to schedule a depart within the resulting window. But it can also happen if
    a tour is very short (e.g. one time period) and the prob spec having a zero probability for
    that tour hour.

    Therefor we need to handle trips that could not be scheduled. There are two ways (at least)
    to solve this problem:

    1) CHOOSE_MOST_INITIAL
    simply assign a depart time to the trip, even if it has a zero probability. It makes
    most sense, in this case, to assign the 'most initial' depart time, so that subsequent trips
    are minimally impacted. This can be done in the final iteration, thus affecting only the
    trips that could no be scheduled by the standard approach

    2) drop_and_cleanup
    drop trips that could no be scheduled, and adjust their leg mates, as is done for failed
    trips in trip_destination.

    For now we are choosing among these approaches with a manifest constant, but this could
    be made a model setting...

    """
    trace_label = "trip_scheduling"

    model_settings = config.read_model_settings('trip_scheduling.yaml')
    assert 'DEPART_ALT_BASE' in model_settings

    failfix = model_settings.get(FAILFIX, FAILFIX_DEFAULT)

    probs_spec = pd.read_csv(config.config_file_path('trip_scheduling_probs.csv'), comment='#')

    trips_df = trips.to_frame()
    tours = tours.to_frame()

    # add tour-based chunk_id so we can chunk all trips in tour together
    trips_df['chunk_id'] = reindex(pd.Series(list(range(len(tours))), tours.index), trips_df.tour_id)

    max_iterations = model_settings.get('MAX_ITERATIONS', 1)
    assert max_iterations > 0

    choices_list = []
    i = 0
    while (i < max_iterations) and not trips_df.empty:

        i += 1
        last_iteration = (i == max_iterations)

        trace_label_i = tracing.extend_trace_label(trace_label, "i%s" % i)
        logger.info("%s scheduling %s trips", trace_label_i, trips_df.shape[0])

        choices = \
            run_trip_scheduling(
                trips_df,
                tours,
                probs_spec,
                model_settings,
                last_iteration=last_iteration,
                trace_hh_id=trace_hh_id,
                chunk_size=chunk_size,
                trace_label=trace_label_i)

        # boolean series of trips whose individual trip scheduling failed
        failed = choices.reindex(trips_df.index).isnull()
        logger.info("%s %s failed", trace_label_i, failed.sum())

        if not last_iteration:
            # boolean series of trips whose leg scheduling failed
            failed_cohorts = failed_trip_cohorts(trips_df, failed)
            trips_df = trips_df[failed_cohorts]
            choices = choices[~failed_cohorts]

        choices_list.append(choices)

    trips_df = trips.to_frame()

    choices = pd.concat(choices_list)
    choices = choices.reindex(trips_df.index)
    if choices.isnull().any():
        logger.warning("%s of %s trips could not be scheduled after %s iterations" %
                       (choices.isnull().sum(), trips_df.shape[0], i))

        if failfix != FAILFIX_DROP_AND_CLEANUP:
            raise RuntimeError("%s setting '%s' not enabled in settings" %
                               (FAILFIX, FAILFIX_DROP_AND_CLEANUP))

        trips_df['failed'] = choices.isnull()
        trips_df = cleanup_failed_trips(trips_df)
        choices = choices.reindex(trips_df.index)

    trips_df['depart'] = choices

    assert not trips_df.depart.isnull().any()

    pipeline.replace_table("trips", trips_df)



# ActivitySim
# See full license in LICENSE.txt.

import sys
import os
import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
logger.addHandler(ch)


inputs = {
    'households': 'final_households.csv',
    'persons': 'final_persons.csv',
    'tours': 'final_tours.csv',
    'joint_tour_participants': 'final_joint_tour_participants.csv',
}

surveys = {
    'households': 'survey_households.csv',
    'persons': 'survey_persons.csv',
    'tours': 'survey_tours.csv',
    'joint_tour_participants': 'survey_joint_tour_participants.csv'
}


args = sys.argv[1:]
assert len(args) == 1, "usage extract_survey_data.py <data_dir>"

data_dir = args[0]

input_dir = os.path.join(data_dir, 'survey_data/')
output_dir = os.path.join(data_dir, 'survey_data/')

configs_dir = os.path.dirname('../example/configs/')

households = pd.read_csv(os.path.join(input_dir, inputs['households']))
persons = pd.read_csv(os.path.join(input_dir, inputs['persons']))
tours = pd.read_csv(os.path.join(input_dir, inputs['tours']))
joint_tour_participants = pd.read_csv(os.path.join(input_dir, inputs['joint_tour_participants']))

households = households[
    ['household_id', 'home_zone_id', 'income', 'hhsize', 'HHT', 'auto_ownership', 'num_workers']
]
persons = persons[
    ['person_id', 'household_id', 'age', 'PNUM', 'sex',
     'pemploy', 'pstudent', 'ptype', 'school_zone_id', 'workplace_zone_id', 'free_parking_at_work']
]
tours = tours[
    ['tour_id', 'person_id', 'household_id', 'tour_type', 'tour_category',
     'destination', 'origin', 'start', 'end', 'tour_mode', 'parent_tour_id']
]
joint_tour_participants = joint_tour_participants[
    ['participant_id', 'tour_id', 'household_id', 'person_id', 'participant_num']
]

households.to_csv(os.path.join(output_dir, surveys['households']), index=False)
persons.to_csv(os.path.join(output_dir, surveys['persons']), index=False)
tours.to_csv(os.path.join(output_dir, surveys['tours']), index=False)
joint_tour_participants.to_csv(os.path.join(output_dir, surveys['joint_tour_participants']), index=False)

# household_id,home_zone_id,income,PERSONS,HHT,VEHICL,workers
raw_households = households[
    ['household_id', 'home_zone_id', 'income', 'hhsize', 'HHT', 'auto_ownership', 'num_workers']
]
raw_households = raw_households.rename({'hhsize': 'PERSONS', 'num_workers': 'workers', 'auto_ownership': 'VEHICL'})
raw_households.to_csv(os.path.join(data_dir, 'households.csv'), index=False)

# person_id,household_id,age,PNUM,sex,pemploy,pstudent,ptype
raw_persons = persons[['person_id', 'household_id', 'age', 'PNUM', 'sex', 'pemploy', 'pstudent', 'ptype']]
raw_persons.to_csv(os.path.join(data_dir, 'persons.csv'), index=False)



# ActivitySim
# See full license in LICENSE.txt.


import pytest
import os
import pandas as pd
import pandas.testing as pdt
from ..tour_frequency import process_mandatory_tours


def mandatory_tour_frequency_alternatives():
    configs_dir = os.path.join(os.path.dirname(__file__), 'configs')
    f = os.path.join(configs_dir, 'mandatory_tour_frequency_alternatives.csv')
    df = pd.read_csv(f, comment='#')
    df.set_index('alt', inplace=True)
    return df


def test_mtf():
    persons = pd.DataFrame({
        "is_worker": [True, True, False, False],
        "mandatory_tour_frequency": ["work1", "work_and_school", "work_and_school", "school2"],
        "school_zone_id": [1, 2, 3, 4],
        "workplace_zone_id": [10, 20, 30, 40],
        "home_zone_id": [100, 200, 300, 400],
        "household_id": [1, 2, 2, 4]
    }, index=[10, 20, 30, 40])

    tour_frequency_alternatives = mandatory_tour_frequency_alternatives()

    mandatory_tours = process_mandatory_tours(persons, tour_frequency_alternatives)

    idx = mandatory_tours.index

    pdt.assert_series_equal(
        mandatory_tours.person_id,
        pd.Series([10, 20, 20, 30, 30, 40, 40], index=idx, name='person_id'))

    pdt.assert_series_equal(
        mandatory_tours.tour_type,
        pd.Series(['work', 'work', 'school', 'work', 'school', 'school', 'school'],
                  index=idx, name='tour_type'))

    # tour_nums for work_and_school non-worker should be flipped
    pdt.assert_series_equal(
        mandatory_tours.tour_num,
        pd.Series([1, 1, 2, 2, 1, 1, 2], index=idx, name='tour_num'))

    pdt.assert_series_equal(
        mandatory_tours.destination,
        pd.Series([10, 20, 2, 30, 3, 4, 4], index=idx, name='destination'))

    pdt.assert_series_equal(
        mandatory_tours.origin,
        pd.Series([100, 200, 200, 300, 300, 400, 400], index=idx, name='origin'))



# ActivitySim
# See full license in LICENSE.txt.

from builtins import range
from builtins import object

import logging

import numpy as np
import pandas as pd

from activitysim.core import pipeline

logger = logging.getLogger(__name__)


# time_window states
# coding for time_window states is chosen so that we can quickly assign a tour
# to windows with a bitwise_or of alt's tdd footprint with existing windows
# e.g. bitwise_or of I_START with I_END equals I_START_END
I_EMPTY = 0x0
I_END = 0x4
I_START = 0x2
I_MIDDLE = 0x7
I_START_END = 0x6

I_BIT_SHIFT = 3

COLLISIONS = [
    [I_START, I_START],
    [I_END, I_END],
    [I_MIDDLE, I_MIDDLE],
    [I_START, I_MIDDLE], [I_MIDDLE, I_START],
    [I_END, I_MIDDLE], [I_MIDDLE, I_END],
    [I_START_END, I_MIDDLE], [I_MIDDLE, I_START_END],
]

COLLISION_LIST = [a + (b << I_BIT_SHIFT) for a, b in COLLISIONS]


# str versions of time windows period states
C_EMPTY = str(I_EMPTY)
C_END = str(I_END)
C_START = str(I_START)
C_MIDDLE = str(I_MIDDLE)
C_START_END = str(I_START_END)


def tour_map(persons, tours, tdd_alts, persons_id_col='person_id'):

    sigil = {
        'empty': '   ',
        'overlap': '+++',
        'work': 'WWW',
        'school': 'SSS',
        'escort': 'esc',
        'shopping': 'shp',
        'othmaint': 'mnt',
        'othdiscr': 'dsc',
        'eatout': 'eat',
        'social': 'soc',
        'eat': 'eat',
        'business': 'bus',
        'maint': 'mnt'
    }

    sigil_type = 'S3'

    # we can only map scheduled tours
    tours = tours[tours.tdd.notnull()]

    # convert tdd_alts_df start, end times to time_windows
    min_period = tdd_alts.start.min()
    max_period = tdd_alts.end.max()
    n_periods = max_period - min_period + 1
    n_persons = len(persons.index)

    agenda = np.array([sigil['empty']]*(n_periods*n_persons), dtype=sigil_type)
    agenda = agenda.reshape(n_persons, n_periods)

    scheduled = np.zeros_like(agenda, dtype=int)
    row_ix_map = pd.Series(list(range(n_persons)), index=persons.index)

    # construct with strings so we can create runs of strings using char * int
    w_strings = [
        '0' * (row.start - min_period) +
        '1' * (row.duration + 1) +
        '0' * (max_period - row.end)
        for idx, row in tdd_alts.iterrows()]

    window_periods = np.asanyarray([list(r) for r in w_strings]).astype(int)
    window_periods_df = pd.DataFrame(data=window_periods, index=tdd_alts.index)

    for keys, nth_tours in tours.groupby(['tour_type', 'tour_type_num'], sort=True):

        tour_type = keys[0]
        tour_sigil = sigil[tour_type]

        # numpy array with one time window row for each row in nth_tours
        tour_windows = window_periods_df.loc[nth_tours.tdd].values

        # row idxs of tour_df group rows in windows
        row_ixs = nth_tours[persons_id_col].map(row_ix_map).values

        # count tours per period
        scheduled[row_ixs] += tour_windows

        # assign tour_char to agenda tour periods
        agenda[row_ixs] = np.where(tour_windows, tour_sigil, agenda[row_ixs])

    # show tour overlaps
    agenda = np.where(scheduled > 1, sigil['overlap'], agenda)

    # a = pd.Series([' '.join(a) for a in agenda], index=persons.index)
    a = pd.DataFrame(data=agenda, columns=[str(w) for w in range(min_period, max_period+1)])

    a.index = persons.index
    a.index.name = persons_id_col

    return a


def create_timetable_windows(rows, tdd_alts):
    """
    create an empty (all available) timetable with one window row per rows.index

    Parameters
    ----------
    rows - pd.DataFrame or Series or orca.DataFrameWrapper
        all we care about is the index
    tdd_alts - pd.DataFrame
        We expect a start and end column, and create a timetable to accomodate all alts
        (with on window of padding at each end)

    so if start is 5 and end is 23, we return something like this:

             4  5  6  7  8  9  10  11  12  13  14  15  16  17  18  19  20  21  22  23  24
    person_id
    30       0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
    109      0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0

    Returns
    -------
    pd.DataFrame indexed by rows.index, and one column of int8 for each time window (plus padding)

    """

    # want something with an index to
    assert rows.index is not None

    # pad windows at both ends of day
    windows = list(range(tdd_alts.start.min() - 1, tdd_alts.end.max() + 2))

    # hdf5 store converts these to strs, se we conform
    window_cols = [str(w) for w in windows]

    UNSCHEDULED = 0

    df = pd.DataFrame(data=UNSCHEDULED,
                      index=rows.index,
                      columns=window_cols,
                      dtype=np.int8)

    return df


class TimeTable(object):
    """
    ::

      tdd_alts_df      tdd_footprints_df
      start  end      '0' '1' '2' '3' '4'...
      5      5    ==>  0   6   0   0   0 ...
      5      6    ==>  0   2   4   0   0 ...
      5      7    ==>  0   2   7   4   0 ...

    """

    def __init__(self, windows_df, tdd_alts_df, table_name=None):
        """

        Parameters
        ----------
        table_name
        time_window_df
        tdd_alts_df
        """

        self.windows_table_name = table_name

        self.windows_df = windows_df
        self.windows = self.windows_df.values
        self.checkpoint_df = None

        # series to map window row index value to window row's ordinal index
        self.window_row_ix = pd.Series(list(range(len(windows_df.index))), index=windows_df.index)

        int_time_periods = [int(c) for c in windows_df.columns.values]
        self.time_ix = pd.Series(list(range(len(windows_df.columns))), index=int_time_periods)

        # - pre-compute window state footprints for every tdd_alt
        min_period = min(int_time_periods)
        max_period = max(int_time_periods)
        # construct with strings so we can create runs of strings using char * int
        w_strings = [
            C_EMPTY * (row.start - min_period) +
            (C_START + C_MIDDLE * (row.duration - 1) if row.duration > 0 else '') +
            (C_END if row.duration > 0 else C_START_END) +
            (C_EMPTY * (max_period - row.end))
            for idx, row in tdd_alts_df.iterrows()]

        # we want range index so we can use raw numpy
        assert (tdd_alts_df.index == list(range(tdd_alts_df.shape[0]))).all()
        self.tdd_footprints = np.asanyarray([list(r) for r in w_strings]).astype(int)

    def begin_transaction(self, transaction_loggers):
        """
        begin a transaction for an estimator or list of estimators
        this permits rolling timetable back to the state at the start of the transaction
        so that timetables can be built for scheduling override choices
        """
        if not isinstance(transaction_loggers, list):
            transaction_loggers = [transaction_loggers]
        for transaction_logger in transaction_loggers:
            transaction_logger.log("timetable.begin_transaction %s" % self.windows_table_name)
        self.checkpoint_df = self.windows_df.copy()
        self.transaction_loggers = transaction_loggers
        pass

    def rollback(self):
        assert self.checkpoint_df is not None
        for logger in self.transaction_loggers:
            logger.log("timetable.rollback %s" % self.windows_table_name)
        self.windows_df = self.checkpoint_df
        self.windows = self.windows_df.values
        self.checkpoint_df = None
        self.transaction_loggers = None

    def slice_windows_by_row_id(self, window_row_ids):
        """
        return windows array slice containing rows for specified window_row_ids
        (in window_row_ids order)
        """
        row_ixs = window_row_ids.map(self.window_row_ix).values
        windows = self.windows[row_ixs]

        return windows

    def slice_windows_by_row_id_and_period(self, window_row_ids, periods):

        # row ixs of tour_df group rows in windows
        row_ixs = window_row_ids.map(self.window_row_ix).values

        # col ixs of periods in windows
        time_col_ixs = periods.map(self.time_ix).values

        windows = self.windows[row_ixs, time_col_ixs]

        return windows

    def get_windows_df(self):

        # It appears that assignments into windows write through to underlying pandas table.
        # because we set windows = windows_df.values, and since all the columns are the same type
        # so no need to refresh pandas dataframe, but if we had to it would go here

        # assert (self.windows_df.values == self.windows).all()
        return self.windows_df

    def replace_table(self):
        """
        Save or replace windows_df  DataFrame to pipeline with saved table name
        (specified when object instantiated.)

        This is a convenience function in case caller instantiates object in one context
        (e.g. dependency injection) where it knows the pipeline table name, but wants to
        checkpoint the table in another context where it does not know that name.
        """

        assert self.windows_table_name is not None
        if self.checkpoint_df is not None:
            for logger in self.transaction_loggers.values():
                logger.log("Attempt to replace_table while in transaction: %s" %
                           self.windows_table_name, level=logging.ERROR)
            raise RuntimeError("Attempt to replace_table while in transaction")

        # get windows_df from bottleneck function in case updates to self.person_window
        # do not write through to pandas dataframe
        pipeline.replace_table(self.windows_table_name, self.get_windows_df())

    def tour_available(self, window_row_ids, tdds):
        """
        test whether time window allows tour with specific tdd alt's time window

        Parameters
        ----------
        window_row_ids : pandas Series
            series of window_row_ids indexed by tour_id
        tdds : pandas series
            series of tdd_alt ids, index irrelevant

        Returns
        -------
        available : pandas Series of bool
            with same index as window_row_ids.index (presumably tour_id, but we don't care)
        """

        assert len(window_row_ids) == len(tdds)

        # numpy array with one tdd_footprints_df row for tdds
        tour_footprints = self.tdd_footprints[tdds.values.astype(int)]

        # numpy array with one windows row for each person
        windows = self.slice_windows_by_row_id(window_row_ids)

        # t0 = tracing.print_elapsed_time("slice_windows_by_row_id", t0, debug=True)

        x = tour_footprints + (windows << I_BIT_SHIFT)

        available = ~np.isin(x, COLLISION_LIST).any(axis=1)
        available = pd.Series(available, index=window_row_ids.index)

        return available

    def assign(self, window_row_ids, tdds):
        """
        Assign tours (represented by tdd alt ids) to persons

        Updates self.windows numpy array. Assignments will not 'take' outside this object
        until/unless replace_table called or updated timetable retrieved by get_windows_df

        Parameters
        ----------
        window_row_ids : pandas Series
            series of window_row_ids indexed by tour_id
        tdds : pandas series
            series of tdd_alt ids, index irrelevant
        """

        assert len(window_row_ids) == len(tdds)

        # vectorization doesn't work duplicates
        assert len(window_row_ids.index) == len(np.unique(window_row_ids.values))

        # numpy array with one time window row for each person tdd
        tour_footprints = self.tdd_footprints[tdds.values.astype(int)]

        # row idxs of windows to assign to
        row_ixs = window_row_ids.map(self.window_row_ix).values

        self.windows[row_ixs] = np.bitwise_or(self.windows[row_ixs], tour_footprints)

    def assign_subtour_mask(self, window_row_ids, tdds):
        """
        ::

          index     window_row_ids   tdds
          20973389  20973389           26
          44612864  44612864            3
          48954854  48954854            7

          tour footprints
          [[0 0 2 7 7 7 7 7 7 4 0 0 0 0 0 0 0 0 0 0 0]
          [0 2 7 7 4 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
          [0 2 7 7 7 7 7 7 4 0 0 0 0 0 0 0 0 0 0 0 0]]

          subtour_mask
          [[7 7 0 0 0 0 0 0 0 0 7 7 7 7 7 7 7 7 7 7 7]
          [7 0 0 0 0 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7]
          [7 0 0 0 0 0 0 0 0 7 7 7 7 7 7 7 7 7 7 7 7]]

        """

        # expect window_row_ids for every row
        assert len(window_row_ids) == len(self.window_row_ix)

        assert len(window_row_ids) == len(tdds)

        self.windows.fill(0)
        self.assign(window_row_ids, tdds)

        # numpy array with one time window row for each person tdd
        tour_footprints = self.tdd_footprints[tdds.values.astype(int)]

        # row idxs of windows to assign to
        row_ixs = window_row_ids.map(self.window_row_ix).values

        self.windows[row_ixs] = (tour_footprints == 0) * I_MIDDLE

    def assign_footprints(self, window_row_ids, footprints):
        """
        assign footprints for specified window_row_ids

        This method is used for initialization of joint_tour timetables based on the
        combined availability of the joint tour participants

        Parameters
        ----------
        window_row_ids : pandas Series
            series of window_row_ids index irrelevant, but we want to use map()
        footprints : numpy array
            with one row per window_row_id and one column per time period
        """

        assert len(window_row_ids) == footprints.shape[0]

        # require same number of periods in footprints
        assert self.windows.shape[1] == footprints.shape[1]

        # vectorization doesn't work with duplicate row_ids
        assert len(window_row_ids.values) == len(np.unique(window_row_ids.values))

        # row idxs of windows to assign to
        row_ixs = window_row_ids.map(self.window_row_ix).values

        self.windows[row_ixs] = np.bitwise_or(self.windows[row_ixs], footprints)

    def pairwise_available(self, window1_row_ids, window2_row_ids):

        available1 = (self.slice_windows_by_row_id(window1_row_ids) != I_MIDDLE) * 1
        available2 = (self.slice_windows_by_row_id(window2_row_ids) != I_MIDDLE) * 1

        return (available1 * available2)

    def individually_available(self, window_row_ids):

        return (self.slice_windows_by_row_id(window_row_ids) != I_MIDDLE) * 1

    def adjacent_window_run_length(self, window_row_ids, periods, before):
        """
        Return the number of adjacent periods before or after specified period
        that are available (not in the middle of another tour.)

        Internal DRY method to implement adjacent_window_before and adjacent_window_after

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant
        before : bool
            Specify desired run length is of adjacent window before (True) or after (False)
        """
        assert len(window_row_ids) == len(periods)

        time_col_ixs = periods.map(self.time_ix).values

        # sliced windows with 1s where windows state is I_MIDDLE and 0s elsewhere
        available = (self.slice_windows_by_row_id(window_row_ids) != I_MIDDLE) * 1

        # padding periods not available
        available[:, 0] = 0
        available[:, -1] = 0

        # column idxs of windows
        num_rows, num_cols = available.shape
        time_col_ix_map = np.tile(np.arange(0, num_cols), num_rows).reshape(num_rows, num_cols)
        # 0 1 2 3 4 5...
        # 0 1 2 3 4 5...
        # 0 1 2 3 4 5...

        if before:
            # ones after specified time, zeroes before
            before_mask = (time_col_ix_map < time_col_ixs.reshape(num_rows, 1)) * 1
            # index of first unavailable window after time
            first_unavailable = np.where((1-available)*before_mask, time_col_ix_map, 0).max(axis=1)
            available_run_length = time_col_ixs - first_unavailable - 1
        else:
            # ones after specified time, zeroes before
            after_mask = (time_col_ix_map > time_col_ixs.reshape(num_rows, 1)) * 1
            # index of first unavailable window after time
            first_unavailable = \
                np.where((1 - available) * after_mask, time_col_ix_map, num_cols).min(axis=1)
            available_run_length = first_unavailable - time_col_ixs - 1

        return pd.Series(available_run_length, index=window_row_ids.index)

    def adjacent_window_before(self, window_row_ids, periods):
        """
        Return number of adjacent periods before specified period that are available
        (not in the middle of another tour.)

        Implements MTC TM1 macro @@getAdjWindowBeforeThisPeriodAlt
        Function name is kind of a misnomer, but parallels that used in MTC TM1 UECs

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant

        Returns
        -------
        pandas Series int
            Number of adjacent windows indexed by window_row_ids.index
        """
        return self.adjacent_window_run_length(window_row_ids, periods, before=True)

    def adjacent_window_after(self, window_row_ids, periods):
        """
        Return number of adjacent periods after specified period that are available
        (not in the middle of another tour.)

        Implements MTC TM1 macro @@adjWindowAfterThisPeriodAlt
        Function name is kind of a misnomer, but parallels that used in MTC TM1 UECs

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant

        Returns
        -------
        pandas Series int
            Number of adjacent windows indexed by window_row_ids.index
        """
        return self.adjacent_window_run_length(window_row_ids, periods, before=False)

    def window_periods_in_states(self, window_row_ids, periods, states):
        """
        Return boolean array indicating whether specified window periods are in list of states.

        Internal DRY method to implement previous_tour_ends and previous_tour_begins

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant (one period per window_row_id)
        states : list of int
            presumably (e.g. I_EMPTY, I_START...)

        Returns
        -------
        pandas Series boolean
            indexed by window_row_ids.index
        """

        assert len(window_row_ids) == len(periods)

        window = self.slice_windows_by_row_id_and_period(window_row_ids, periods)

        return pd.Series(np.isin(window, states), window_row_ids.index)

    def previous_tour_ends(self, window_row_ids, periods):
        """
        Does a previously scheduled tour end in the specified period?

        Implements MTC TM1 @@prevTourEndsThisDeparturePeriodAlt

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant (one period per window_row_id)

        Returns
        -------
        pandas Series boolean
            indexed by window_row_ids.index
        """
        return self.window_periods_in_states(window_row_ids, periods, [I_END, I_START_END])

    def previous_tour_begins(self, window_row_ids, periods):
        """
        Does a previously scheduled tour begin in the specified period?

        Implements MTC TM1 @@prevTourBeginsThisArrivalPeriodAlt

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        periods : pandas series int
            series of tdd_alt ids, index irrelevant

        Returns
        -------
        pandas Series boolean
            indexed by window_row_ids.index
        """

        return self.window_periods_in_states(window_row_ids, periods, [I_START, I_START_END])

    def remaining_periods_available(self, window_row_ids, starts, ends):
        """
        Determine number of periods remaining available after the time window from starts to ends
        is hypothetically scheduled

        Implements MTC TM1 @@remainingPeriodsAvailableAlt

        The start and end periods will always be available after scheduling, so ignore them.
        The periods between start and end must be currently unscheduled, so assume they will become
        unavailable after scheduling this window.

        Parameters
        ----------
        window_row_ids : pandas Series int
            series of window_row_ids indexed by tour_id
        starts : pandas series int
            series of tdd_alt ids, index irrelevant (one per window_row_id)
        ends : pandas series int
            series of tdd_alt ids, index irrelevant (one per window_row_id)

        Returns
        -------
        available : pandas Series int
            number periods available indexed by window_row_ids.index
        """

        assert len(window_row_ids) == len(starts)
        assert len(window_row_ids) == len(ends)

        available = (self.slice_windows_by_row_id(window_row_ids) != I_MIDDLE).sum(axis=1)

        # don't count time window padding at both ends of day
        available -= 2

        available -= np.clip((ends - starts - 1), a_min=0, a_max=None)
        available = pd.Series(available, index=window_row_ids.index)

        return available



# ActivitySim
# See full license in LICENSE.txt.

import os
import shutil

import logging

import yaml

import pandas as pd

from activitysim.core import config
from activitysim.core import simulate

from activitysim.core.util import reindex

logger = logging.getLogger('estimation')

ESTIMATION_SETTINGS_FILE_NAME = 'estimation.yaml'


class Estimator(object):

    def __init__(self, model_name, settings_name, estimation_table_recipes):

        logger.info("Initialize Estimator for'%s'" % (model_name,))

        self.model_name = model_name
        self.settings_name = settings_name
        self.estimation_table_recipes = estimation_table_recipes
        self.estimating = True

        # ensure the output data directory exists
        output_dir = self.data_directory()
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)  # make directory if needed

        # delete estimation files
        file_type = ('csv', 'yaml')
        for file_name in os.listdir(output_dir):
            if file_name.startswith(model_name) and file_name.endswith(file_type):
                file_path = os.path.join(output_dir, file_name)
                try:
                    if os.path.isfile(file_path):
                        os.unlink(file_path)
                except Exception as e:
                    print(e)

        # FIXME - not required?
        # assert 'override_choices' in self.model_settings, \
        #     "override_choices not found for %s in %s." % (model_name, ESTIMATION_SETTINGS_FILE_NAME)

        self.omnibus_tables = self.estimation_table_recipes['omnibus_tables']
        self.omnibus_tables_append_columns = self.estimation_table_recipes['omnibus_tables_append_columns']
        self.tables = {}
        self.tables_to_cache = [table_name for tables in self.omnibus_tables.values() for table_name in tables]
        self.alt_id_column_name = None
        self.chooser_id_column_name = None

    def log(self, msg, level=logging.INFO):
        logger.log(level, "%s: %s" % (self.model_name, msg))

    def info(self, msg):
        self.log(msg, level=logging.INFO)

    def debug(self, msg):
        self.log(msg, level=logging.DEBUG)

    def warning(self, msg):
        self.log(msg, level=logging.WARNING)

    def set_alt_id(self, alt_id):
        self.alt_id_column_name = alt_id

    def get_alt_id(self):
        if self.alt_id_column_name is None:
            self.warning("alt_id is None. Did you forget to call set_alt_id()?")
        assert self.alt_id_column_name is not None
        return self.alt_id_column_name

    def set_chooser_id(self, chooser_id_column_name):
        self.chooser_id_column_name = chooser_id_column_name

    def get_chooser_id(self):
        if self.chooser_id_column_name is None:
            self.warning("chooser_id is None. Did you forget to call set_chooser_id()?")
        assert self.chooser_id_column_name is not None
        return self.chooser_id_column_name

    def end_estimation(self):

        self.write_omnibus_table()

        self.estimating = False
        self.tables = None

        self.info("end estimation")

        manager.release(self)

    def data_directory(self):

        # shouldn't be asking for this if not estimating
        assert self.estimating
        assert self.settings_name is not None

        parent_dir = config.output_file_path('estimation_data_bundle')

        if self.settings_name != self.model_name:
            parent_dir = os.path.join(parent_dir, self.settings_name)

        return os.path.join(parent_dir, self.model_name)

    def file_path(self, table_name, file_type=None):

        # shouldn't be asking for this if not estimating
        assert self.estimating

        if file_type:
            file_name = "%s_%s.%s" % (self.model_name, table_name, file_type)
        else:
            file_name = "%s_%s" % (self.model_name, table_name)

        return os.path.join(self.data_directory(), file_name)

    def write_table(self, df, table_name, index=True, append=True):

        def cache_table(df, table_name, append):
            if table_name in self.tables and not append:
                raise RuntimeError("cache_table %s append=False and table exists" % (table_name,))
            if table_name in self.tables:
                self.tables[table_name] = pd.concat([self.tables[table_name], df])
            else:
                self.tables[table_name] = df.copy()

        def write_table(df, table_name, index, append):
            file_path = self.file_path(table_name, 'csv')
            file_exists = os.path.isfile(file_path)
            if file_exists and not append:
                raise RuntimeError("write_table %s append=False and file exists: %s" % (table_name, file_path))
            df.to_csv(file_path, mode='a', index=index, header=(not file_exists))

        assert self.estimating

        cache = table_name in self.tables_to_cache
        write = not cache
        # write = True

        if cache:
            cache_table(df, table_name, append)
            self.debug('write_table cache: %s' % table_name)

        if write:
            write_table(df, table_name, index, append)
            self.debug('write_table write: %s' % table_name)

    def write_omnibus_table(self):

        if len(self.omnibus_tables) == 0:
            return

        for omnibus_table, table_names in self.omnibus_tables.items():

            self.debug("write_omnibus_table: %s table_names: %s" % (omnibus_table, table_names))
            for t in table_names:
                if t not in self.tables:
                    self.warning("write_omnibus_table: %s table '%s' not found" % (omnibus_table, t))

            # ignore any ables not in cache
            table_names = [t for t in table_names if t in self.tables]
            concat_axis = 1 if omnibus_table in self.omnibus_tables_append_columns else 0

            df = pd.concat([self.tables[t] for t in table_names], axis=concat_axis)

            file_path = self.file_path(omnibus_table, 'csv')

            assert not os.path.isfile(file_path)

            df.sort_index(ascending=True, inplace=True, kind='mergesort')
            df.to_csv(file_path, mode='a', index=True, header=True)

            self.debug('write_omnibus_choosers: %s' % file_path)

    def write_dict(self, d, dict_name):

        assert self.estimating

        file_path = self.file_path(dict_name, 'yaml')

        # we don't know how to concat, and afraid to overwrite
        assert not os.path.isfile(file_path)

        with open(file_path, 'w') as f:
            # write ordered dict as array
            yaml.dump(d, f)

        self.debug("estimate.write_dict: %s" % file_path)

    def write_coefficients(self, coefficients_df, tag='coefficients'):
        assert self.estimating
        self.write_table(coefficients_df, tag, append=False)

    def write_coefficients_template(self, coefficients_df, tag='coefficients_template'):
        assert self.estimating
        self.write_table(coefficients_df, tag, append=False)

    def write_choosers(self, choosers_df):
        self.write_table(choosers_df, 'choosers', append=True)

    def write_choices(self, choices):
        if isinstance(choices, pd.Series):
            choices = choices.to_frame(name='model_choice')
        assert(list(choices.columns) == ['model_choice'])
        self.write_table(choices, 'choices', append=True)

    def write_override_choices(self, choices):
        if isinstance(choices, pd.Series):
            choices = choices.to_frame(name='override_choice')
        assert(list(choices.columns) == ['override_choice'])
        self.write_table(choices, 'override_choices', append=True)

    def write_constants(self, constants):
        self.write_dict(self, constants, 'model_constants')

    def write_nest_spec(self, nest_spec):
        self.write_dict(self, nest_spec, 'nest_spec')

    def copy_model_settings(self, settings_file_name, tag='model_settings'):

        input_path = config.base_settings_file_path(settings_file_name)

        output_path = self.file_path(tag, 'yaml')

        shutil.copy(input_path, output_path)

    def write_model_settings(self, model_settings, settings_file_name):

        self.copy_model_settings(settings_file_name)
        if 'inherit_settings' in model_settings:
            self.write_dict(model_settings, 'inherited_model_settings')

    def melt_alternatives(self, df):

        alt_id_name = self.alt_id_column_name

        assert alt_id_name is not None, \
            "alt_id not set. Did you forget to call set_alt_id()? (%s)" % self.model_name
        assert alt_id_name in df, \
            "alt_id_column_name '%s' not in alternatives table (%s)" % (alt_id_name, self.model_name)

        variable_column = 'variable'

        #            alt_dest  util_dist_0_1  util_dist_1_2  ...
        # person_id                                          ...
        # 31153             1            1.0           0.75  ...
        # 31153             2            1.0           0.46  ...
        # 31153             3            1.0           0.28  ...

        if df.index.name is not None:
            chooser_name = df.index.name
            assert self.chooser_id_column_name in (chooser_name, None)
            df = df.reset_index()
        else:
            assert self.chooser_id_column_name is not None
            chooser_name = self.chooser_id_column_name
            assert chooser_name in df

        # mergesort is the only stable sort, and we want the expressions to appear in original df column order
        melt_df = pd.melt(df, id_vars=[chooser_name, alt_id_name]) \
            .sort_values(by=chooser_name, kind='mergesort') \
            .rename(columns={'variable': variable_column})

        # person_id,alt_dest,expression,value
        # 31153,1,util_dist_0_1,1.0
        # 31153,2,util_dist_0_1,1.0
        # 31153,3,util_dist_0_1,1.0

        melt_df = melt_df.set_index([chooser_name, variable_column, alt_id_name]).unstack(2)
        melt_df.columns = melt_df.columns.droplevel(0)
        melt_df = melt_df.reset_index(1)

        # person_id,expression,1,2,3,4,5,...
        # 31153,util_dist_0_1,0.75,0.46,0.27,0.63,0.48,...
        # 31153,util_dist_1_2,0.0,0.0,0.0,0.0,0.0,...
        # 31153,util_dist_2_3,0.0,0.0,0.0,0.0,0.0,...

        return melt_df

    def write_interaction_expression_values(self, df):
        df = self.melt_alternatives(df)
        self.write_table(df, 'interaction_expression_values', append=True)

    def write_expression_values(self, df):
        self.write_table(df, 'expression_values', append=True)

    def write_alternatives(self, alternatives_df):
        self.write_table(alternatives_df, 'alternatives', append=True)

    def write_interaction_sample_alternatives(self, alternatives_df):
        alternatives_df = self.melt_alternatives(alternatives_df)
        self.write_table(alternatives_df, 'interaction_sample_alternatives', append=True)

    def write_interaction_simulate_alternatives(self, interaction_df):
        interaction_df = self.melt_alternatives(interaction_df)
        self.write_table(interaction_df, 'interaction_simulate_alternatives', append=True)

    def join_survey_values(self, df, table_name, left_on=None, right_on=None):
        # convenience method so deep callers don't need to import estimation
        assert self.estimating
        return manager.join_survey_values(df, table_name, left_on, right_on)

    def get_survey_values(self, model_values, table_name, column_names):
        # convenience method so deep callers don't need to import estimation
        assert self.estimating
        return manager.get_survey_values(model_values, table_name, column_names)

    def write_spec(self, model_settings=None, file_name=None, tag='SPEC'):

        if model_settings is not None:
            assert file_name is None
            file_name = model_settings[tag]

        input_path = config.config_file_path(file_name)
        output_path = self.file_path(table_name=tag, file_type='csv')
        shutil.copy(input_path, output_path)
        self.debug("estimate.write_spec: %s" % output_path)


class EstimationManager(object):

    def __init__(self):

        self.settings_initialized = False
        self.bundles = []
        self.estimation_table_recipes = {}
        self.model_estimation_table_types = {}
        self.estimating = {}

    def initialize_settings(self):

        assert not self.settings_initialized
        settings = config.read_model_settings(ESTIMATION_SETTINGS_FILE_NAME)
        self.enabled = settings.get('enable', 'True')
        self.bundles = settings.get('bundles', [])
        self.model_estimation_table_types = settings.get('model_estimation_table_types', {})
        self.estimation_table_recipes = settings.get('estimation_table_recipes', {})

        if self.enabled:
            self.survey_tables = settings.get('survey_tables', {})
            for table_name, table_info in self.survey_tables.items():
                assert 'file_name' in table_info, \
                    "No file name specified for survey_table '%s' in %s" % (table_name, ESTIMATION_SETTINGS_FILE_NAME)
                file_path = config.data_file_path(table_info['file_name'], mandatory=True)
                assert os.path.exists(file_path), \
                    "File for survey table '%s' not found: %s" % (table_name, file_path)
                df = pd.read_csv(file_path)
                index_col = table_info.get('index_col')
                if index_col is not None:
                    assert index_col in df.columns, \
                        "Index col '%s' not in survey_table '%s' in file: %s % (index_col, table_name, file_path)"
                    df.set_index(index_col, inplace=True)

                # add the table df to survey_tables
                table_info['df'] = df

        self.settings_initialized = True

    def begin_estimation(self, model_name, bundle_name=None):
        """
        begin estimating of model_name is specified as model to estimate, otherwise return False

        Parameters
        ----------
        model_name

        Returns
        -------

        """
        # load estimation settings file
        if not self.settings_initialized:
            self.initialize_settings()

        # global estimation setting
        if not self.enabled:
            return None

        bundle_name = bundle_name or model_name

        if bundle_name not in self.bundles:
            return None

        # can't estimate the same model simultaneously
        assert model_name not in self.estimating, \
            "Cant begin estimating %s - already estimating that model." % (model_name, )

        assert bundle_name in self.model_estimation_table_types, \
            "No estimation_table_type for %s in %s." % (bundle_name, ESTIMATION_SETTINGS_FILE_NAME)

        model_estimation_table_type = self.model_estimation_table_types[bundle_name]

        assert model_estimation_table_type in self.estimation_table_recipes, \
            "model_estimation_table_type '%s' for model %s no in %s." % \
            (model_estimation_table_type, model_name, ESTIMATION_SETTINGS_FILE_NAME)

        self.estimating[model_name] = \
            Estimator(model_name, model_name,
                      estimation_table_recipes=self.estimation_table_recipes[model_estimation_table_type])

        return self.estimating[model_name]

    def release(self, estimator):

        self.estimating.pop(estimator.model_name)

    def get_survey_table(self, table_name):
        assert self.enabled
        if table_name not in self.survey_tables:
            logger.warning("EstimationManager. get_survey_table: survey table '%s' not in survey_tables" % table_name)
        return self.survey_tables[table_name].get('df')

    def get_survey_values(self, model_values, table_name, column_names):

        assert isinstance(model_values, (pd.Series, pd.DataFrame, pd.Index)), \
            "get_survey_values model_values has unrecognized type %s" % type(model_values)

        dest_index = model_values if isinstance(model_values, (pd.Index)) else model_values.index

        # read override_df table
        survey_df = manager.get_survey_table(table_name)

        assert survey_df is not None, \
            "get_survey_values: table '%s' not found" % (table_name,)

        column_name = column_names if isinstance(column_names, str) else None
        if column_name:
            column_names = [column_name]

        if not set(column_names).issubset(set(survey_df.columns)):
            missing_columns = list(set(column_names) - set(survey_df.columns))
            logger.error("missing columns (%s) in survey table %s" % (missing_columns, table_name))
            print("survey table columns: %s" % (survey_df.columns, ))
            raise RuntimeError("missing columns (%s) in survey table %s" % (missing_columns, table_name))

        assert set(column_names).issubset(set(survey_df.columns)), \
            "missing columns (%s) in survey table %s" % (list(set(column_names) - set(survey_df.columns)), table_name)

        # for now tour_id is asim_tour_id in survey_df
        asim_df_index_name = dest_index.name
        if asim_df_index_name == survey_df.index.name:
            # survey table has same index as activitysim
            survey_df_index_column = 'index'
        elif asim_df_index_name in survey_df.columns:
            # survey table has activitysim index as column
            survey_df_index_column = asim_df_index_name
        elif 'asim_%s' % asim_df_index_name in survey_df.columns:
            # survey table has activitysim index as column with asim_ prefix
            survey_df_index_column = 'asim_%s' % asim_df_index_name
        else:
            logger.error("get_survey_values:index '%s' not in survey table" % dest_index.name)
            # raise RuntimeError("index '%s' not in survey table %s" % (dest_index.name, table_name)
            survey_df_index_column = None

        logger.debug("get_survey_values: reindexing using %s.%s" % (table_name, survey_df_index_column))

        values = pd.DataFrame(index=dest_index)
        for c in column_names:
            if survey_df_index_column == 'index':
                survey_values = survey_df[c]
            else:
                survey_values = pd.Series(survey_df[c].values, index=survey_df[survey_df_index_column])

            survey_values = reindex(survey_values, dest_index)

            # shouldn't be any choices we can't override
            missing_values = survey_values.isna()
            if missing_values.any():
                logger.error("missing survey_values for %s\n%s" % (c, dest_index[missing_values]))
                logger.error("couldn't get_survey_values for %s in %s\n" % (c, table_name))
                raise RuntimeError("couldn't get_survey_values for %s in %s\n" % (c, table_name))

            values[c] = survey_values

        return values[column_name] if column_name else values

    def join_survey_values(self, model_df, table_name, left_on, right_on):

        left_on = left_on or list(model_df.columns)
        right_on = right_on or left_on

        # read survey_df table
        survey_df = manager.get_survey_table(table_name)
        assert survey_df is not None, "estimation.get_override_choices: table '%s' not found" % (table_name,)

        assert isinstance(model_df, pd.DataFrame), \
            "I expected model_df to be a DataFrame, but got %s" % (type(model_df),)

        assert right_on is not None, "I expected to find 'right_on' column list in merge settings."
        assert set(right_on).issubset(set(survey_df.columns)), \
            "Not all 'right_on' columns (%s) not found in %s" % (right_on, table_name)

        df = pd.merge(model_df.reset_index(), survey_df[right_on],
                      left_on=left_on, right_on=right_on, how='left').set_index(model_df.index.name)
        return df


manager = EstimationManager()



# Orca
# Copyright (C) 2016 UrbanSim Inc.
# See full license in LICENSE.

import os
import tempfile

import tables
import pandas as pd
import pytest
from pandas.util import testing as pdt

from activitysim.core import orca
from activitysim.core import inject
from .utils_testing import assert_frames_equal


def setup_function(func):
    orca.clear_all()
    orca.enable_cache()


def teardown_function(func):
    orca.clear_all()
    orca.enable_cache()
    # be nice to the others tests that expect decorated injectables to be there
    inject.reinject_decorated_tables()


@pytest.fixture
def df():
    return pd.DataFrame(
        [[1, 4],
         [2, 5],
         [3, 6]],
        columns=['a', 'b'],
        index=['x', 'y', 'z'])


def test_tables(df):
    wrapped_df = orca.add_table('test_frame', df)

    @orca.table()
    def test_func(test_frame):
        return test_frame.to_frame() / 2

    assert set(orca.list_tables()) == {'test_frame', 'test_func'}

    table = orca.get_table('test_frame')
    assert table is wrapped_df
    assert table.columns == ['a', 'b']
    assert table.local_columns == ['a', 'b']
    assert len(table) == 3
    pdt.assert_index_equal(table.index, df.index)
    pdt.assert_series_equal(table.get_column('a'), df.a)
    pdt.assert_series_equal(table.a, df.a)
    pdt.assert_series_equal(table['b'], df['b'])

    table = orca._TABLES['test_func']
    assert table.index is None
    assert table.columns == []
    assert len(table) == 0
    pdt.assert_frame_equal(table.to_frame(), df / 2)
    pdt.assert_frame_equal(table.to_frame([]), df[[]])
    pdt.assert_frame_equal(table.to_frame(columns=['a']), df[['a']] / 2)
    pdt.assert_frame_equal(table.to_frame(columns='a'), df[['a']] / 2)
    pdt.assert_index_equal(table.index, df.index)
    pdt.assert_series_equal(table.get_column('a'), df.a / 2)
    pdt.assert_series_equal(table.a, df.a / 2)
    pdt.assert_series_equal(table['b'], df['b'] / 2)
    assert len(table) == 3
    assert table.columns == ['a', 'b']


def test_table_func_cache(df):
    orca.add_injectable('x', 2)

    @orca.table(cache=True)
    def table(variable='x'):
        return df * variable

    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 2)
    orca.add_injectable('x', 3)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 2)
    orca.get_table('table').clear_cached()
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 3)
    orca.add_injectable('x', 4)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 3)
    orca.clear_cache()
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 4)
    orca.add_injectable('x', 5)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 4)
    orca.add_table('table', table)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 5)


def test_table_func_cache_disabled(df):
    orca.add_injectable('x', 2)

    @orca.table('table', cache=True)
    def asdf(x):
        return df * x

    orca.disable_cache()

    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 2)
    orca.add_injectable('x', 3)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 3)

    orca.enable_cache()

    orca.add_injectable('x', 4)
    pdt.assert_frame_equal(orca.get_table('table').to_frame(), df * 3)


def test_table_copy(df):
    orca.add_table('test_frame_copied', df, copy_col=True)
    orca.add_table('test_frame_uncopied', df, copy_col=False)
    orca.add_table('test_func_copied', lambda: df, copy_col=True)
    orca.add_table('test_func_uncopied', lambda: df, copy_col=False)

    @orca.table(copy_col=True)
    def test_funcd_copied():
        return df

    @orca.table(copy_col=False)
    def test_funcd_uncopied():
        return df

    @orca.table(copy_col=True)
    def test_funcd_copied2(test_frame_copied):
        # local returns original, but it is copied by copy_col.
        return test_frame_copied.local

    @orca.table(copy_col=True)
    def test_funcd_copied3(test_frame_uncopied):
        # local returns original, but it is copied by copy_col.
        return test_frame_uncopied.local

    @orca.table(copy_col=False)
    def test_funcd_uncopied2(test_frame_copied):
        # local returns original.
        return test_frame_copied.local

    @orca.table(copy_col=False)
    def test_funcd_uncopied3(test_frame_uncopied):
        # local returns original.
        return test_frame_uncopied.local

    orca.add_table('test_cache_copied', lambda: df, cache=True, copy_col=True)
    orca.add_table(
        'test_cache_uncopied', lambda: df, cache=True, copy_col=False)

    @orca.table(cache=True, copy_col=True)
    def test_cached_copied():
        return df

    @orca.table(cache=True, copy_col=False)
    def test_cached_uncopied():
        return df

    # Create tables with computed columns.
    orca.add_table(
        'test_copied_columns', pd.DataFrame(index=df.index), copy_col=True)
    orca.add_table(
        'test_uncopied_columns', pd.DataFrame(index=df.index), copy_col=False)
    for column_name in ['a', 'b']:
        label = "test_frame_uncopied.{}".format(column_name)

        def func(col=label):
            return col
        for table_name in ['test_copied_columns', 'test_uncopied_columns']:
            orca.add_column(table_name, column_name, func)

    for name in ['test_frame_uncopied', 'test_func_uncopied',
                 'test_funcd_uncopied', 'test_funcd_uncopied2',
                 'test_funcd_uncopied3', 'test_cache_uncopied',
                 'test_cached_uncopied', 'test_uncopied_columns',
                 'test_frame_copied', 'test_func_copied',
                 'test_funcd_copied', 'test_funcd_copied2',
                 'test_funcd_copied3', 'test_cache_copied',
                 'test_cached_copied', 'test_copied_columns']:
        table = orca.get_table(name)
        table2 = orca.get_table(name)

        # to_frame will always return a copy.
        if 'columns' in name:
            assert_frames_equal(table.to_frame(), df)
        else:
            pdt.assert_frame_equal(table.to_frame(), df)
        assert table.to_frame() is not df
        pdt.assert_frame_equal(table.to_frame(), table.to_frame())
        assert table.to_frame() is not table.to_frame()
        pdt.assert_series_equal(table.to_frame()['a'], df['a'])
        assert table.to_frame()['a'] is not df['a']
        pdt.assert_series_equal(table.to_frame()['a'],
                                table.to_frame()['a'])
        assert table.to_frame()['a'] is not table.to_frame()['a']

        if 'uncopied' in name:
            pdt.assert_series_equal(table['a'], df['a'])
            assert table['a'] is df['a']
            pdt.assert_series_equal(table['a'], table2['a'])
            assert table['a'] is table2['a']
        else:
            pdt.assert_series_equal(table['a'], df['a'])
            assert table['a'] is not df['a']
            pdt.assert_series_equal(table['a'], table2['a'])
            assert table['a'] is not table2['a']


def test_columns_for_table():
    orca.add_column(
        'table1', 'col10', pd.Series([1, 2, 3], index=['a', 'b', 'c']))
    orca.add_column(
        'table2', 'col20', pd.Series([10, 11, 12], index=['x', 'y', 'z']))

    @orca.column('table1')
    def col11():
        return pd.Series([4, 5, 6], index=['a', 'b', 'c'])

    @orca.column('table2', 'col21')
    def asdf():
        return pd.Series([13, 14, 15], index=['x', 'y', 'z'])

    t1_col_names = orca.list_columns_for_table('table1')
    assert set(t1_col_names) == {'col10', 'col11'}

    t2_col_names = orca.list_columns_for_table('table2')
    assert set(t2_col_names) == {'col20', 'col21'}

    t1_cols = orca._columns_for_table('table1')
    assert 'col10' in t1_cols and 'col11' in t1_cols

    t2_cols = orca._columns_for_table('table2')
    assert 'col20' in t2_cols and 'col21' in t2_cols


def test_columns_and_tables(df):
    orca.add_table('test_frame', df)

    @orca.table()
    def test_func(test_frame):
        return test_frame.to_frame() / 2

    orca.add_column('test_frame', 'c', pd.Series([7, 8, 9], index=df.index))

    @orca.column('test_func', 'd')
    def asdf(test_func):
        return test_func.to_frame(columns=['b'])['b'] * 2

    @orca.column('test_func')
    def e(column='test_func.d'):
        return column + 1

    test_frame = orca.get_table('test_frame')
    assert set(test_frame.columns) == set(['a', 'b', 'c'])
    assert_frames_equal(
        test_frame.to_frame(),
        pd.DataFrame(
            {'a': [1, 2, 3],
             'b': [4, 5, 6],
             'c': [7, 8, 9]},
            index=['x', 'y', 'z']))
    assert_frames_equal(
        test_frame.to_frame(columns=['a', 'c']),
        pd.DataFrame(
            {'a': [1, 2, 3],
             'c': [7, 8, 9]},
            index=['x', 'y', 'z']))

    test_func_df = orca._TABLES['test_func']
    assert set(test_func_df.columns) == set(['d', 'e'])
    assert_frames_equal(
        test_func_df.to_frame(),
        pd.DataFrame(
            {'a': [0.5, 1, 1.5],
             'b': [2, 2.5, 3],
             'c': [3.5, 4, 4.5],
             'd': [4., 5., 6.],
             'e': [5., 6., 7.]},
            index=['x', 'y', 'z']))
    assert_frames_equal(
        test_func_df.to_frame(columns=['b', 'd']),
        pd.DataFrame(
            {'b': [2, 2.5, 3],
             'd': [4., 5., 6.]},
            index=['x', 'y', 'z']))
    assert set(test_func_df.columns) == set(['a', 'b', 'c', 'd', 'e'])

    assert set(orca.list_columns()) == {
        ('test_frame', 'c'), ('test_func', 'd'), ('test_func', 'e')}


def test_column_cache(df):
    orca.add_injectable('x', 2)
    series = pd.Series([1, 2, 3], index=['x', 'y', 'z'])
    key = ('table', 'col')

    @orca.table()
    def table():
        return df

    @orca.column(*key, cache=True)
    def column(variable='x'):
        return series * variable

    def c():
        return orca._COLUMNS[key]

    pdt.assert_series_equal(c()(), series * 2)
    orca.add_injectable('x', 3)
    pdt.assert_series_equal(c()(), series * 2)
    c().clear_cached()
    pdt.assert_series_equal(c()(), series * 3)
    orca.add_injectable('x', 4)
    pdt.assert_series_equal(c()(), series * 3)
    orca.clear_cache()
    pdt.assert_series_equal(c()(), series * 4)
    orca.add_injectable('x', 5)
    pdt.assert_series_equal(c()(), series * 4)
    orca.get_table('table').clear_cached()
    pdt.assert_series_equal(c()(), series * 5)
    orca.add_injectable('x', 6)
    pdt.assert_series_equal(c()(), series * 5)
    orca.add_column(*key, column=column, cache=True)
    pdt.assert_series_equal(c()(), series * 6)


def test_column_cache_disabled(df):
    orca.add_injectable('x', 2)
    series = pd.Series([1, 2, 3], index=['x', 'y', 'z'])
    key = ('table', 'col')

    @orca.table()
    def table():
        return df

    @orca.column(*key, cache=True)
    def column(x):
        return series * x

    def c():
        return orca._COLUMNS[key]

    orca.disable_cache()

    pdt.assert_series_equal(c()(), series * 2)
    orca.add_injectable('x', 3)
    pdt.assert_series_equal(c()(), series * 3)

    orca.enable_cache()

    orca.add_injectable('x', 4)
    pdt.assert_series_equal(c()(), series * 3)


def test_update_col(df):
    wrapped = orca.add_table('table', df)

    wrapped.update_col('b', pd.Series([7, 8, 9], index=df.index))
    pdt.assert_series_equal(
        wrapped['b'], pd.Series([7, 8, 9], index=df.index, name='b'))

    a_dtype = wrapped['a'].dtype

    # test 1 - cast the data type before the update
    wrapped.update_col_from_series('a', pd.Series(dtype=a_dtype))
    pdt.assert_series_equal(wrapped['a'], df['a'])

    # test 2 - let the update method do the cast
    wrapped.update_col_from_series('a', pd.Series(dtype='float64'), True)
    pdt.assert_series_equal(wrapped['a'], df['a'])

    # test 3 - don't cast, should raise an error
    with pytest.raises(ValueError):
        wrapped.update_col_from_series('a', pd.Series(dtype='float64'))

    wrapped.update_col_from_series('a', pd.Series([99], index=['y']))
    pdt.assert_series_equal(
        wrapped['a'], pd.Series([1, 99, 3], index=df.index, name='a'))


class _FakeTable(object):
    def __init__(self, name, columns):
        self.name = name
        self.columns = columns


@pytest.fixture
def fta():
    return _FakeTable('a', ['aa', 'ab', 'ac'])


@pytest.fixture
def ftb():
    return _FakeTable('b', ['bx', 'by', 'bz'])


def test_column_map_raises(fta, ftb):
    with pytest.raises(RuntimeError):
        orca.column_map([fta, ftb], ['aa', 'by', 'bz', 'cw'])


def test_column_map_none(fta, ftb):
    assert orca.column_map([fta, ftb], None) == {'a': None, 'b': None}


def test_column_map(fta, ftb):
    result = orca.column_map([fta, ftb], ['aa', 'by', 'bz'])
    assert result['a'] == ['aa']
    assert sorted(result['b']) == ['by', 'bz']

    result = orca.column_map([fta, ftb], ['by', 'bz'])
    assert result['a'] == []
    assert sorted(result['b']) == ['by', 'bz']


def test_is_step():
    @orca.step()
    def test_step():
        pass

    assert orca.is_step('test_step') is True
    assert orca.is_step('not_a_step') is False


def test_steps(df):
    orca.add_table('test_table', df)

    df2 = df / 2
    orca.add_table('test_table2', df2)

    @orca.step()
    def test_step(test_table, test_column='test_table2.b'):
        tt = test_table.to_frame()
        test_table['a'] = tt['a'] + tt['b']
        pdt.assert_series_equal(test_column, df2['b'])

    with pytest.raises(KeyError):
        orca.get_step('asdf')

    step = orca.get_step('test_step')
    assert step._tables_used() == set(['test_table', 'test_table2'])
    step()

    table = orca.get_table('test_table')
    pdt.assert_frame_equal(
        table.to_frame(),
        pd.DataFrame(
            {'a': [5, 7, 9],
             'b': [4, 5, 6]},
            index=['x', 'y', 'z']))

    assert orca.list_steps() == ['test_step']


def test_step_run(df):
    orca.add_table('test_table', df)

    @orca.table()
    def table_func(test_table):
        tt = test_table.to_frame()
        tt['c'] = [7, 8, 9]
        return tt

    @orca.column('table_func')
    def new_col(test_table, table_func):
        tt = test_table.to_frame()
        tf = table_func.to_frame(columns=['c'])
        return tt['a'] + tt['b'] + tf['c']

    @orca.step()
    def test_step1(iter_var, test_table, table_func):
        tf = table_func.to_frame(columns=['new_col'])
        test_table[iter_var] = tf['new_col'] + iter_var

    @orca.step('test_step2')
    def asdf(table='test_table'):
        tt = table.to_frame()
        table['a'] = tt['a'] ** 2

    orca.run(steps=['test_step1', 'test_step2'], iter_vars=[2000, 3000])

    test_table = orca.get_table('test_table')
    assert_frames_equal(
        test_table.to_frame(),
        pd.DataFrame(
            {'a': [1, 16, 81],
             'b': [4, 5, 6],
             2000: [2012, 2015, 2018],
             3000: [3012, 3017, 3024]},
            index=['x', 'y', 'z']))

    m = orca.get_step('test_step1')
    assert set(m._tables_used()) == {'test_table', 'table_func'}


def test_step_func_source_data():
    @orca.step()
    def test_step():
        return 'orca'

    filename, lineno, source = orca.get_step('test_step').func_source_data()

    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert source == (
        "    @orca.step()\n"
        "    def test_step():\n"
        "        return 'orca'\n")


def test_get_broadcast():
    orca.broadcast('a', 'b', cast_on='ax', onto_on='bx')
    orca.broadcast('x', 'y', cast_on='yx', onto_index=True)

    assert orca.is_broadcast('a', 'b') is True
    assert orca.is_broadcast('b', 'a') is False

    with pytest.raises(KeyError):
        orca.get_broadcast('b', 'a')

    ab = orca.get_broadcast('a', 'b')
    assert isinstance(ab, orca.Broadcast)
    assert ab == ('a', 'b', 'ax', 'bx', False, False)

    xy = orca.get_broadcast('x', 'y')
    assert isinstance(xy, orca.Broadcast)
    assert xy == ('x', 'y', 'yx', None, False, True)


def test_get_broadcasts():
    orca.broadcast('a', 'b')
    orca.broadcast('b', 'c')
    orca.broadcast('z', 'b')
    orca.broadcast('f', 'g')

    with pytest.raises(ValueError):
        orca._get_broadcasts(['a', 'b', 'g'])

    assert set(orca._get_broadcasts(['a', 'b', 'c', 'z']).keys()) == \
        {('a', 'b'), ('b', 'c'), ('z', 'b')}
    assert set(orca._get_broadcasts(['a', 'b', 'z']).keys()) == \
        {('a', 'b'), ('z', 'b')}
    assert set(orca._get_broadcasts(['a', 'b', 'c']).keys()) == \
        {('a', 'b'), ('b', 'c')}

    assert set(orca.list_broadcasts()) == \
        {('a', 'b'), ('b', 'c'), ('z', 'b'), ('f', 'g')}


def test_collect_variables(df):
    orca.add_table('df', df)

    @orca.table()
    def df_func():
        return df

    @orca.column('df')
    def zzz():
        return df['a'] / 2

    orca.add_injectable('answer', 42)

    @orca.injectable()
    def injected():
        return 'injected'

    @orca.table('source table', cache=True)
    def source():
        return df

    with pytest.raises(KeyError):
        orca._collect_variables(['asdf'])

    with pytest.raises(KeyError):
        orca._collect_variables(names=['df'], expressions=['asdf'])

    names = ['df', 'df_func', 'answer', 'injected', 'source_label', 'df_a']
    expressions = ['source table', 'df.a']
    things = orca._collect_variables(names, expressions)

    assert set(things.keys()) == set(names)
    assert isinstance(things['source_label'], orca.DataFrameWrapper)
    pdt.assert_frame_equal(things['source_label'].to_frame(), df)
    assert isinstance(things['df_a'], pd.Series)
    pdt.assert_series_equal(things['df_a'], df['a'])


def test_collect_variables_expression_only(df):
    @orca.table()
    def table():
        return df

    vars = orca._collect_variables(['a'], ['table.a'])
    pdt.assert_series_equal(vars['a'], df.a)


def test_injectables():
    orca.add_injectable('answer', 42)

    @orca.injectable()
    def func1(answer):
        return answer * 2

    @orca.injectable('func2', autocall=False)
    def asdf(variable='x'):
        return variable / 2

    @orca.injectable()
    def func3(func2):
        return func2(4)

    @orca.injectable()
    def func4(func='func1'):
        return func / 2

    assert orca._INJECTABLES['answer'] == 42
    assert orca._INJECTABLES['func1']() == 42 * 2
    assert orca._INJECTABLES['func2'](4) == 2
    assert orca._INJECTABLES['func3']() == 2
    assert orca._INJECTABLES['func4']() == 42

    assert orca.get_injectable('answer') == 42
    assert orca.get_injectable('func1') == 42 * 2
    assert orca.get_injectable('func2')(4) == 2
    assert orca.get_injectable('func3') == 2
    assert orca.get_injectable('func4') == 42

    with pytest.raises(KeyError):
        orca.get_injectable('asdf')

    assert set(orca.list_injectables()) == \
        {'answer', 'func1', 'func2', 'func3', 'func4'}


def test_injectables_combined(df):
    @orca.injectable()
    def column():
        return pd.Series(['a', 'b', 'c'], index=df.index)

    @orca.table()
    def table():
        return df

    @orca.step()
    def step(table, column):
        df = table.to_frame()
        df['new'] = column
        orca.add_table('table', df)

    orca.run(steps=['step'])

    table_wr = orca.get_table('table').to_frame()

    pdt.assert_frame_equal(table_wr[['a', 'b']], df)
    pdt.assert_series_equal(table_wr['new'], pd.Series(column(), name='new'))


def test_injectables_cache():
    x = 2

    @orca.injectable(autocall=True, cache=True)
    def inj():
        return x * x

    def i():
        return orca._INJECTABLES['inj']

    assert i()() == 4
    x = 3
    assert i()() == 4
    i().clear_cached()
    assert i()() == 9
    x = 4
    assert i()() == 9
    orca.clear_cache()
    assert i()() == 16
    x = 5
    assert i()() == 16
    orca.add_injectable('inj', inj, autocall=True, cache=True)
    assert i()() == 25


def test_injectables_cache_disabled():
    x = 2

    @orca.injectable(autocall=True, cache=True)
    def inj():
        return x * x

    def i():
        return orca._INJECTABLES['inj']

    orca.disable_cache()

    assert i()() == 4
    x = 3
    assert i()() == 9

    orca.enable_cache()

    assert i()() == 9
    x = 4
    assert i()() == 9

    orca.disable_cache()
    assert i()() == 16


def test_memoized_injectable():
    outside = 'x'

    @orca.injectable(autocall=False, memoize=True)
    def x(s):
        return outside + s

    assert 'x' in orca._MEMOIZED

    def getx():
        return orca.get_injectable('x')

    assert hasattr(getx(), 'cache')
    assert hasattr(getx(), 'clear_cached')

    assert getx()('y') == 'xy'
    outside = 'z'
    assert getx()('y') == 'xy'

    getx().clear_cached()

    assert getx()('y') == 'zy'


def test_memoized_injectable_cache_off():
    outside = 'x'

    @orca.injectable(autocall=False, memoize=True)
    def x(s):
        return outside + s

    def getx():
        return orca.get_injectable('x')('y')

    orca.disable_cache()

    assert getx() == 'xy'
    outside = 'z'
    assert getx() == 'zy'

    orca.enable_cache()
    outside = 'a'

    assert getx() == 'zy'

    orca.disable_cache()

    assert getx() == 'ay'


def test_clear_cache_all(df):
    @orca.table(cache=True)
    def table():
        return df

    @orca.column('table', cache=True)
    def z(table):
        return df.a

    @orca.injectable(cache=True)
    def x():
        return 'x'

    @orca.injectable(autocall=False, memoize=True)
    def y(s):
        return s + 'y'

    orca.eval_variable('table.z')
    orca.eval_variable('x')
    orca.get_injectable('y')('x')

    assert list(orca._TABLE_CACHE.keys()) == ['table']
    assert list(orca._COLUMN_CACHE.keys()) == [('table', 'z')]
    assert list(orca._INJECTABLE_CACHE.keys()) == ['x']
    assert orca._MEMOIZED['y'].value.cache == {(('x',), None): 'xy'}

    orca.clear_cache()

    assert orca._TABLE_CACHE == {}
    assert orca._COLUMN_CACHE == {}
    assert orca._INJECTABLE_CACHE == {}
    assert orca._MEMOIZED['y'].value.cache == {}


def test_clear_cache_scopes(df):
    @orca.table(cache=True, cache_scope='forever')
    def table():
        return df

    @orca.column('table', cache=True, cache_scope='iteration')
    def z(table):
        return df.a

    @orca.injectable(cache=True, cache_scope='step')
    def x():
        return 'x'

    @orca.injectable(autocall=False, memoize=True, cache_scope='iteration')
    def y(s):
        return s + 'y'

    orca.eval_variable('table.z')
    orca.eval_variable('x')
    orca.get_injectable('y')('x')

    assert list(orca._TABLE_CACHE.keys()) == ['table']
    assert list(orca._COLUMN_CACHE.keys()) == [('table', 'z')]
    assert list(orca._INJECTABLE_CACHE.keys()) == ['x']
    assert orca._MEMOIZED['y'].value.cache == {(('x',), None): 'xy'}

    orca.clear_cache(scope='step')

    assert list(orca._TABLE_CACHE.keys()) == ['table']
    assert list(orca._COLUMN_CACHE.keys()) == [('table', 'z')]
    assert orca._INJECTABLE_CACHE == {}
    assert orca._MEMOIZED['y'].value.cache == {(('x',), None): 'xy'}

    orca.clear_cache(scope='iteration')

    assert list(orca._TABLE_CACHE.keys()) == ['table']
    assert orca._COLUMN_CACHE == {}
    assert orca._INJECTABLE_CACHE == {}
    assert orca._MEMOIZED['y'].value.cache == {}

    orca.clear_cache(scope='forever')

    assert orca._TABLE_CACHE == {}
    assert orca._COLUMN_CACHE == {}
    assert orca._INJECTABLE_CACHE == {}
    assert orca._MEMOIZED['y'].value.cache == {}


def test_cache_scope(df):
    orca.add_injectable('x', 11)
    orca.add_injectable('y', 22)
    orca.add_injectable('z', 33)
    orca.add_injectable('iterations', 1)

    @orca.injectable(cache=True, cache_scope='forever')
    def a(x):
        return x

    @orca.injectable(cache=True, cache_scope='iteration')
    def b(y):
        return y

    @orca.injectable(cache=True, cache_scope='step')
    def c(z):
        return z

    @orca.step()
    def m1(iter_var, a, b, c):
        orca.add_injectable('x', iter_var + a)
        orca.add_injectable('y', iter_var + b)
        orca.add_injectable('z', iter_var + c)

        assert a == 11

    @orca.step()
    def m2(iter_var, a, b, c, iterations):
        assert a == 11
        if iter_var == 1000:
            assert b == 22
            assert c == 1033
        elif iter_var == 2000:
            assert b == 1022
            assert c == 3033

        orca.add_injectable('iterations', iterations + 1)

    orca.run(['m1', 'm2'], iter_vars=[1000, 2000])


def test_table_func_local_cols(df):
    @orca.table()
    def table():
        return df
    orca.add_column(
        'table', 'new', pd.Series(['a', 'b', 'c'], index=df.index))

    assert orca.get_table('table').local_columns == ['a', 'b']


def test_is_table(df):
    orca.add_table('table', df)
    assert orca.is_table('table') is True
    assert orca.is_table('asdf') is False


@pytest.fixture
def store_name(request):
    fname = tempfile.NamedTemporaryFile(suffix='.h5').name

    def fin():
        if os.path.isfile(fname):
            os.remove(fname)
    request.addfinalizer(fin)

    return fname


@pytest.mark.filterwarnings('ignore::tables.NaturalNameWarning')
def test_write_tables(df, store_name):
    orca.add_table('table', df)

    @orca.step()
    def step(table):
        pass

    step_tables = orca.get_step_table_names(['step'])

    orca.write_tables(store_name, step_tables)
    with pd.HDFStore(store_name, mode='r') as store:
        assert 'table' in store
        pdt.assert_frame_equal(store['table'], df)

    orca.write_tables(store_name, step_tables, prefix=1969)

    with pd.HDFStore(store_name, mode='r') as store:
        assert '1969/table' in store
        pdt.assert_frame_equal(store['1969/table'], df)


def test_write_all_tables(df, store_name):
    orca.add_table('table', df)
    orca.write_tables(store_name)

    with pd.HDFStore(store_name, mode='r') as store:
        for t in orca.list_tables():
            assert t in store


@pytest.mark.filterwarnings('ignore::tables.NaturalNameWarning')
def test_run_and_write_tables(df, store_name):
    orca.add_table('table', df)

    def year_key(y):
        return '{}'.format(y)

    def series_year(y):
        return pd.Series([y] * 3, index=df.index, name=str(y))

    @orca.step()
    def step(iter_var, table):
        table[year_key(iter_var)] = series_year(iter_var)

    orca.run(
        ['step'], iter_vars=range(11), data_out=store_name, out_interval=3)

    with pd.HDFStore(store_name, mode='r') as store:
        for year in range(0, 11, 3):
            key = '{}/table'.format(year)
            assert key in store

            for x in range(year):
                pdt.assert_series_equal(
                    store[key][year_key(x)], series_year(x))

        assert 'base/table' in store

        for x in range(11):
            pdt.assert_series_equal(
                store['10/table'][year_key(x)], series_year(x))


@pytest.mark.filterwarnings('ignore::tables.NaturalNameWarning')
def test_run_and_write_tables_out_tables_provided(df, store_name):
    table_names = ['table', 'table2', 'table3']
    for t in table_names:
        orca.add_table(t, df)

    @orca.step()
    def step(iter_var, table, table2):
        return

    orca.run(
        ['step'],
        iter_vars=range(1),
        data_out=store_name,
        out_base_tables=table_names,
        out_run_tables=['table'])

    with pd.HDFStore(store_name, mode='r') as store:

        for t in table_names:
            assert 'base/{}'.format(t) in store

        assert '0/table' in store
        assert '0/table2' not in store
        assert '0/table3' not in store


def test_get_raw_table(df):
    orca.add_table('table1', df)

    @orca.table()
    def table2():
        return df

    assert isinstance(orca.get_raw_table('table1'), orca.DataFrameWrapper)
    assert isinstance(orca.get_raw_table('table2'), orca.TableFuncWrapper)

    assert orca.table_type('table1') == 'dataframe'
    assert orca.table_type('table2') == 'function'


def test_get_table(df):
    orca.add_table('frame', df)

    @orca.table()
    def table():
        return df

    @orca.table(cache=True)
    def source():
        return df

    fr = orca.get_table('frame')
    ta = orca.get_table('table')
    so = orca.get_table('source')

    with pytest.raises(KeyError):
        orca.get_table('asdf')

    assert isinstance(fr, orca.DataFrameWrapper)
    assert isinstance(ta, orca.DataFrameWrapper)
    assert isinstance(so, orca.DataFrameWrapper)

    pdt.assert_frame_equal(fr.to_frame(), df)
    pdt.assert_frame_equal(ta.to_frame(), df)
    pdt.assert_frame_equal(so.to_frame(), df)


def test_cache_disabled_cm():
    x = 3

    @orca.injectable(cache=True)
    def xi():
        return x

    assert orca.get_injectable('xi') == 3
    x = 5
    assert orca.get_injectable('xi') == 3

    with orca.cache_disabled():
        assert orca.get_injectable('xi') == 5

    # cache still gets updated even when cacheing is off
    assert orca.get_injectable('xi') == 5


def test_injectables_cm():
    orca.add_injectable('a', 'a')
    orca.add_injectable('b', 'b')
    orca.add_injectable('c', 'c')

    with orca.injectables():
        assert orca._INJECTABLES == {
            'a': 'a', 'b': 'b', 'c': 'c'
        }

    with orca.injectables(c='d', x='x', y='y', z='z'):
        assert orca._INJECTABLES == {
            'a': 'a', 'b': 'b', 'c': 'd',
            'x': 'x', 'y': 'y', 'z': 'z'
        }

    assert orca._INJECTABLES == {
        'a': 'a', 'b': 'b', 'c': 'c'
    }


def test_temporary_tables_cm():
    orca.add_table('a', pd.DataFrame())

    with orca.temporary_tables():
        assert sorted(orca._TABLES.keys()) == ['a']

    with orca.temporary_tables(a=pd.DataFrame(), b=pd.DataFrame()):
        assert sorted(orca._TABLES.keys()) == ['a', 'b']

    assert sorted(orca._TABLES.keys()) == ['a']


def test_is_expression():
    assert orca.is_expression('name') is False
    assert orca.is_expression('table.column') is True


def test_eval_variable(df):
    orca.add_injectable('x', 3)
    assert orca.eval_variable('x') == 3

    @orca.injectable()
    def func(x):
        return 'xyz' * x
    assert orca.eval_variable('func') == 'xyzxyzxyz'
    assert orca.eval_variable('func', x=2) == 'xyzxyz'

    @orca.table()
    def table(x):
        return df * x
    pdt.assert_series_equal(orca.eval_variable('table.a'), df.a * 3)


def test_eval_step(df):
    orca.add_injectable('x', 3)

    @orca.step()
    def step(x):
        return df * x

    pdt.assert_frame_equal(orca.eval_step('step'), df * 3)
    pdt.assert_frame_equal(orca.eval_step('step', x=5), df * 5)


def test_always_dataframewrapper(df):
    @orca.table()
    def table():
        return df / 2

    @orca.table()
    def table2(table):
        assert isinstance(table, orca.DataFrameWrapper)
        return table.to_frame() / 2

    result = orca.eval_variable('table2')
    pdt.assert_frame_equal(result.to_frame(), df / 4)


def test_table_func_source_data(df):
    @orca.table()
    def table():
        return df * 2

    t = orca.get_raw_table('table')
    filename, lineno, source = t.func_source_data()

    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert 'return df * 2' in source


def test_column_type(df):
    orca.add_table('test_frame', df)

    @orca.table()
    def test_func():
        return df

    s = pd.Series(range(len(df)), index=df.index)

    def col_func():
        return s

    orca.add_column('test_frame', 'col_series', s)
    orca.add_column('test_func', 'col_series', s)
    orca.add_column('test_frame', 'col_func', col_func)
    orca.add_column('test_func', 'col_func', col_func)

    tframe = orca.get_raw_table('test_frame')
    tfunc = orca.get_raw_table('test_func')

    assert tframe.column_type('a') == 'local'
    assert tframe.column_type('col_series') == 'series'
    assert tframe.column_type('col_func') == 'function'

    assert tfunc.column_type('a') == 'local'
    assert tfunc.column_type('col_series') == 'series'
    assert tfunc.column_type('col_func') == 'function'


def test_get_raw_column(df):
    orca.add_table('test_frame', df)

    s = pd.Series(range(len(df)), index=df.index)

    def col_func():
        return s

    orca.add_column('test_frame', 'col_series', s)
    orca.add_column('test_frame', 'col_func', col_func)

    assert isinstance(
        orca.get_raw_column('test_frame', 'col_series'),
        orca._SeriesWrapper)
    assert isinstance(
        orca.get_raw_column('test_frame', 'col_func'),
        orca._ColumnFuncWrapper)


def test_column_func_source_data(df):
    orca.add_table('test_frame', df)

    @orca.column('test_frame')
    def col_func():
        return pd.Series(range(len(df)), index=df.index)

    s = orca.get_raw_column('test_frame', 'col_func')
    filename, lineno, source = s.func_source_data()

    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert 'def col_func():' in source


def test_is_injectable():
    orca.add_injectable('answer', 42)
    assert orca.is_injectable('answer') is True
    assert orca.is_injectable('nope') is False


def test_injectable_type():
    orca.add_injectable('answer', 42)

    @orca.injectable()
    def inj1():
        return 42

    @orca.injectable(autocall=False, memoize=True)
    def power(x):
        return 42 ** x

    assert orca.injectable_type('answer') == 'variable'
    assert orca.injectable_type('inj1') == 'function'
    assert orca.injectable_type('power') == 'function'


def test_get_injectable_func_source_data():
    @orca.injectable()
    def inj1():
        return 42

    @orca.injectable(autocall=False, memoize=True)
    def power(x):
        return 42 ** x

    def inj2():
        return 'orca'

    orca.add_injectable('inj2', inj2, autocall=False)

    filename, lineno, source = orca.get_injectable_func_source_data('inj1')
    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert '@orca.injectable()' in source

    filename, lineno, source = orca.get_injectable_func_source_data('power')
    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert '@orca.injectable(autocall=False, memoize=True)' in source

    filename, lineno, source = orca.get_injectable_func_source_data('inj2')
    assert filename.endswith('test_orca.py')
    assert isinstance(lineno, int)
    assert 'def inj2()' in source



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import simulate
from activitysim.core import tracing
from activitysim.core import pipeline
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import expressions

from .util import estimation

from .util.overlap import hh_time_window_overlap


logger = logging.getLogger(__name__)


def add_null_results(trace_label, tours):
    logger.info("Skipping %s: add_null_results" % trace_label)
    tours['composition'] = ''
    pipeline.replace_table("tours", tours)


@inject.step()
def joint_tour_composition(
        tours, households, persons,
        chunk_size,
        trace_hh_id):
    """
    This model predicts the makeup of the travel party (adults, children, or mixed).
    """
    trace_label = 'joint_tour_composition'
    model_settings_file_name = 'joint_tour_composition.yaml'

    tours = tours.to_frame()
    joint_tours = tours[tours.tour_category == 'joint']

    # - if no joint tours
    if joint_tours.shape[0] == 0:
        add_null_results(trace_label, tours)
        return

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation('joint_tour_composition')

    # - only interested in households with joint_tours
    households = households.to_frame()
    households = households[households.num_hh_joint_tours > 0]

    persons = persons.to_frame()
    persons = persons[persons.household_id.isin(households.index)]

    logger.info("Running joint_tour_composition with %d joint tours" % joint_tours.shape[0])

    # - run preprocessor
    preprocessor_settings = model_settings.get('preprocessor', None)
    if preprocessor_settings:

        locals_dict = {
            'persons': persons,
            'hh_time_window_overlap': hh_time_window_overlap
        }

        expressions.assign_columns(
            df=households,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label)

    joint_tours_merged = pd.merge(joint_tours, households,
                                  left_on='household_id', right_index=True, how='left')

    # - simple_simulate
    model_spec = simulate.read_model_spec(file_name=model_settings['SPEC'])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    if estimator:
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_coefficients(coefficients_df)
        estimator.write_choosers(joint_tours_merged)

    choices = simulate.simple_simulate(
        choosers=joint_tours_merged,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name='composition',
        estimator=estimator)

    # convert indexes to alternative names
    choices = pd.Series(model_spec.columns[choices.values], index=choices.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, 'tours', 'composition')
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # add composition column to tours for tracing
    joint_tours['composition'] = choices

    # reindex since we ran model on a subset of households
    tours['composition'] = choices.reindex(tours.index).fillna('').astype(str)
    pipeline.replace_table("tours", tours)

    tracing.print_summary('joint_tour_composition', joint_tours.composition,
                          value_counts=True)

    if trace_hh_id:
        tracing.trace_df(joint_tours,
                         label="joint_tour_composition.joint_tours",
                         slicer='household_id')



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd
import numpy as np

from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import simulate
from activitysim.core.mem import force_garbage_collect
from activitysim.core.util import assign_in_place

from activitysim.core import los
from activitysim.core.pathbuilder import TransitVirtualPathBuilder

from .util.mode import run_tour_mode_choice_simulate
from .util import estimation

logger = logging.getLogger(__name__)

"""
Tour mode choice is run for all tours to determine the transportation mode that
will be used for the tour
"""


def write_coefficient_template(model_settings):
    coefficients = simulate.read_model_coefficients(model_settings)

    coefficients = coefficients.transpose()
    coefficients.columns.name = None

    template = coefficients.copy()

    coef_names = []
    coef_values = []

    for c in coefficients.columns:

        values = coefficients[c]
        unique_values = values.unique()

        for uv in unique_values:

            if len(unique_values) == 1:
                uv_coef_name = c + '_all'
            else:
                uv_coef_name = c + '_' + '_'.join(values[values == uv].index.values)

            coef_names.append(uv_coef_name)
            coef_values.append(uv)

            template[c] = template[c].where(values != uv, uv_coef_name)

    refactored_coefficients = pd.DataFrame({'coefficient_name': coef_names,  'value': coef_values})
    refactored_coefficients.value = refactored_coefficients.value.astype(np.float32)
    print(refactored_coefficients)

    template = template.transpose()
    template.to_csv(
        config.output_file_path('tour_mode_choice_coefficients_template.csv'),
        mode='w', index=True, header=True)

    refactored_coefficients.to_csv(
        config.output_file_path('tour_mode_choice_refactored_coefficients.csv'),
        mode='w', index=False, header=True)


@inject.step()
def tour_mode_choice_simulate(tours, persons_merged,
                              network_los,
                              chunk_size,
                              trace_hh_id):
    """
    Tour mode choice simulate
    """
    trace_label = 'tour_mode_choice'
    model_settings_file_name = 'tour_mode_choice.yaml'
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get('MODE_CHOICE_LOGSUM_COLUMN_NAME')
    mode_column_name = 'tour_mode'  # FIXME - should be passed in?

    primary_tours = tours.to_frame()
    assert not (primary_tours.tour_category == 'atwork').any()

    logger.info("Running %s with %d tours" % (trace_label, primary_tours.shape[0]))

    tracing.print_summary('tour_types',
                          primary_tours.tour_type, value_counts=True)

    persons_merged = persons_merged.to_frame()
    primary_tours_merged = pd.merge(primary_tours, persons_merged, left_on='person_id',
                                    right_index=True, how='left', suffixes=('', '_r'))

    constants = {}
    # model_constants can appear in expressions
    constants.update(config.get_model_constants(model_settings))

    skim_dict = network_los.get_default_skim_dict()

    # setup skim keys
    orig_col_name = 'home_zone_id'
    dest_col_name = 'destination'

    out_time_col_name = 'start'
    in_time_col_name = 'end'
    odt_skim_stack_wrapper = skim_dict.wrap_3d(orig_key=orig_col_name, dest_key=dest_col_name,
                                               dim3_key='out_period')
    dot_skim_stack_wrapper = skim_dict.wrap_3d(orig_key=dest_col_name, dest_key=orig_col_name,
                                               dim3_key='in_period')
    odr_skim_stack_wrapper = skim_dict.wrap_3d(orig_key=orig_col_name, dest_key=dest_col_name,
                                               dim3_key='in_period')
    dor_skim_stack_wrapper = skim_dict.wrap_3d(orig_key=dest_col_name, dest_key=orig_col_name,
                                               dim3_key='out_period')
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,  # dot return skims for e.g. TNC bridge return fare
        "dor_skims": dor_skim_stack_wrapper,  # odt return skims for e.g. TNC bridge return fare
        "od_skims": od_skim_stack_wrapper,
        'orig_col_name': orig_col_name,
        'dest_col_name': dest_col_name,
        'out_time_col_name': out_time_col_name,
        'in_time_col_name': in_time_col_name
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?

        tvpb = network_los.tvpb

        tvpb_logsum_odt = tvpb.wrap_logsum(orig_key=orig_col_name, dest_key=dest_col_name,
                                           tod_key='out_period', segment_key='demographic_segment',
                                           cache_choices=True,
                                           trace_label=trace_label, tag='tvpb_logsum_odt')
        tvpb_logsum_dot = tvpb.wrap_logsum(orig_key=dest_col_name, dest_key=orig_col_name,
                                           tod_key='in_period', segment_key='demographic_segment',
                                           cache_choices=True,
                                           trace_label=trace_label, tag='tvpb_logsum_dot')

        skims.update({
            'tvpb_logsum_odt': tvpb_logsum_odt,
            'tvpb_logsum_dot': tvpb_logsum_dot
        })

        # TVPB constants can appear in expressions
        constants.update(network_los.setting('TVPB_SETTINGS.tour_mode_choice.CONSTANTS'))

    estimator = estimation.manager.begin_estimation('tour_mode_choice')
    if estimator:
        estimator.write_coefficients(simulate.read_model_coefficients(model_settings))
        estimator.write_coefficients_template(simulate.read_model_coefficient_template(model_settings))
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # (run_tour_mode_choice_simulate writes choosers post-annotation)

    # FIXME should normalize handling of tour_type and tour_purpose
    # mtctm1 school tour_type includes univ, which has different coefficients from elementary and HS
    # we should either add this column when tours created or add univ to tour_types
    not_university = (primary_tours_merged.tour_type != 'school') | ~primary_tours_merged.is_university
    primary_tours_merged['tour_purpose'] = \
        primary_tours_merged.tour_type.where(not_university, 'univ')

    choices_list = []
    for tour_purpose, tours_segment in primary_tours_merged.groupby('tour_purpose'):

        logger.info("tour_mode_choice_simulate tour_type '%s' (%s tours)" %
                    (tour_purpose, len(tours_segment.index), ))

        if network_los.zone_system == los.THREE_ZONE:
            tvpb_logsum_odt.extend_trace_label(tour_purpose)
            tvpb_logsum_dot.extend_trace_label(tour_purpose)

        # name index so tracing knows how to slice
        assert tours_segment.index.name == 'tour_id'

        choices_df = run_tour_mode_choice_simulate(
            tours_segment,
            tour_purpose, model_settings,
            mode_column_name=mode_column_name,
            logsum_column_name=logsum_column_name,
            network_los=network_los,
            skims=skims,
            constants=constants,
            estimator=estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(trace_label, tour_purpose),
            trace_choice_name='tour_mode_choice')

        tracing.print_summary('tour_mode_choice_simulate %s choices_df' % tour_purpose,
                              choices_df.tour_mode, value_counts=True)

        choices_list.append(choices_df)

        # FIXME - force garbage collection
        force_garbage_collect()

    choices_df = pd.concat(choices_list)

    # add cached tvpb_logsum tap choices for modes specified in tvpb_mode_path_types
    if network_los.zone_system == los.THREE_ZONE:

        tvpb_mode_path_types = model_settings.get('tvpb_mode_path_types')
        for mode, path_types in tvpb_mode_path_types.items():

            for direction, skim in zip(['od', 'do'], [tvpb_logsum_odt, tvpb_logsum_dot]):

                path_type = path_types[direction]
                skim_cache = skim.cache[path_type]

                print(f"mode {mode} direction {direction} path_type {path_type}")

                for c in skim_cache:

                    dest_col = f'{direction}_{c}'

                    if dest_col not in choices_df:
                        choices_df[dest_col] = 0 if pd.api.types.is_numeric_dtype(skim_cache[c]) else ''
                    choices_df[dest_col].where(choices_df.tour_mode != mode, skim_cache[c], inplace=True)

    if estimator:
        estimator.write_choices(choices_df.tour_mode)
        choices_df.tour_mode = estimator.get_survey_values(choices_df.tour_mode, 'tours', 'tour_mode')
        estimator.write_override_choices(choices_df.tour_mode)
        estimator.end_estimation()

    tracing.print_summary('tour_mode_choice_simulate all tour type choices',
                          choices_df.tour_mode, value_counts=True)

    # so we can trace with annotations
    assign_in_place(primary_tours, choices_df)

    # update tours table with mode choice (and optionally logsums)
    all_tours = tours.to_frame()
    assign_in_place(all_tours, choices_df)

    pipeline.replace_table("tours", all_tours)

    if trace_hh_id:
        tracing.trace_df(primary_tours,
                         label=tracing.extend_trace_label(trace_label, mode_column_name),
                         slicer='tour_id',
                         index_label='tour_id',
                         warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import pipeline
from activitysim.core import inject
from activitysim.core import tracing
from activitysim.core import mem

from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


def read_raw_persons(households):

    df = read_input_table("persons")

    if inject.get_injectable('households_sliced', False):
        # keep only persons in the sampled households
        df = df[df.household_id.isin(households.index)]

    return df


@inject.table()
def persons(households, trace_hh_id):

    df = read_raw_persons(households)

    logger.info("loaded persons %s" % (df.shape,))

    # replace table function with dataframe
    inject.add_table('persons', df)

    pipeline.get_rn_generator().add_channel('persons', df)

    tracing.register_traceable_table('persons', df)
    if trace_hh_id:
        tracing.trace_df(df, "raw.persons", warn_if_empty=True)

    print(f"{len(df.household_id.unique())} unique household_ids in persons")
    print(f"{len(households.index.unique())} unique household_ids in households")
    assert not households.index.duplicated().any()
    assert not df.index.duplicated().any()

    persons_without_households = ~df.household_id.isin(households.index)
    if persons_without_households.any():
        logger.error(f"{persons_without_households.sum()} persons out of {len(persons)} without households\n"
                     f"{pd.Series({'person_id': persons_without_households.index.values})}")
        raise RuntimeError(f"{persons_without_households.sum()} persons with bad household_id")

    households_without_persons = df.groupby('household_id').size().reindex(households.index).isnull()
    if households_without_persons.any():
        logger.error(f"{households_without_persons.sum()} households out of {len(households.index)} without  persons\n"
                     f"{pd.Series({'household_id': households_without_persons.index.values})}")
        raise RuntimeError(f"{households_without_persons.sum()} households with no persons")

    return df


# another common merge for persons
@inject.table()
def persons_merged(persons, households, land_use, accessibility):

    return inject.merge_tables(persons.name, tables=[persons, households, land_use, accessibility])



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd
import numpy as np

from activitysim.core import simulate
from activitysim.core import tracing
from activitysim.core import pipeline
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import timetable as tt
from activitysim.core import expressions

from .util.vectorize_tour_scheduling import vectorize_subtour_scheduling

from .util import estimation

from activitysim.core.util import assign_in_place

logger = logging.getLogger(__name__)

DUMP = False


@inject.step()
def atwork_subtour_scheduling(
        tours,
        persons_merged,
        tdd_alts,
        skim_dict,
        chunk_size,
        trace_hh_id):
    """
    This model predicts the departure time and duration of each activity for at work subtours tours
    """

    trace_label = 'atwork_subtour_scheduling'
    model_settings_file_name = 'tour_scheduling_atwork.yaml'

    tours = tours.to_frame()
    subtours = tours[tours.tour_category == 'atwork']

    # - if no atwork subtours
    if subtours.shape[0] == 0:
        tracing.no_results(trace_label)
        return

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation('atwork_subtour_scheduling')

    model_spec = simulate.read_model_spec(file_name=model_settings['SPEC'])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    persons_merged = persons_merged.to_frame()

    logger.info("Running %s with %d tours", trace_label, len(subtours))

    # preprocessor
    constants = config.get_model_constants(model_settings)
    od_skim_wrapper = skim_dict.wrap('origin', 'destination')
    skims = {
        "od_skims": od_skim_wrapper,
    }
    expressions.annotate_preprocessors(
        subtours, constants, skims,
        model_settings, trace_label)

    # parent_tours table with columns ['tour_id', 'tdd'] index = tour_id
    parent_tour_ids = subtours.parent_tour_id.astype(np.int64).unique()
    parent_tours = pd.DataFrame({'tour_id': parent_tour_ids}, index=parent_tour_ids)
    parent_tours = parent_tours.merge(tours[['tdd']], left_index=True, right_index=True)

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df)
        # we don't need to update timetable because subtours are scheduled inside work trip windows

    choices = vectorize_subtour_scheduling(
        parent_tours,
        subtours,
        persons_merged,
        tdd_alts, model_spec,
        model_settings,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_label=trace_label)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, 'tours', 'tdd')
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # choices are tdd alternative ids
    # we want to add start, end, and duration columns to tours, which we have in tdd_alts table
    tdd_choices = pd.merge(choices.to_frame('tdd'), tdd_alts, left_on=['tdd'], right_index=True, how='left')

    assign_in_place(tours, tdd_choices)
    pipeline.replace_table("tours", tours)

    if trace_hh_id:
        tracing.trace_df(tours[tours.tour_category == 'atwork'],
                         label="atwork_subtour_scheduling",
                         slicer='person_id',
                         index_label='tour_id',
                         columns=None)

    if DUMP:
        subtours = tours[tours.tour_category == 'atwork']
        parent_tours = tours[tours.index.isin(subtours.parent_tour_id)]

        tracing.dump_df(DUMP, subtours, trace_label, 'sub_tours')
        tracing.dump_df(DUMP, parent_tours, trace_label, 'parent_tours')

        parent_tours['parent_tour_id'] = parent_tours.index
        subtours = pd.concat([parent_tours, subtours])
        tracing.dump_df(DUMP,
                        tt.tour_map(parent_tours, subtours, tdd_alts,
                                    persons_id_col='parent_tour_id'),
                        trace_label, 'tour_map')



import os
import sys
import requests
import shutil
import glob
import pkg_resources
import yaml

PACKAGE = 'activitysim'
EXAMPLES_DIR = 'examples'
MANIFEST = 'example_manifest.yaml'


def _example_path(resource):
    resource = os.path.join(EXAMPLES_DIR, resource)
    path = pkg_resources.resource_filename(PACKAGE, resource)

    return path


def _load_manifest():
    with open(_example_path(MANIFEST), 'r') as f:
        manifest = yaml.safe_load(f.read())

    assert manifest, f'error: could not load {MANIFEST}'
    return {example['name']: example for example in manifest}


EXAMPLES = _load_manifest()


def add_create_args(parser):
    """Create command args
    """
    create_group = parser.add_mutually_exclusive_group(required=True)
    create_group.add_argument('-l', '--list',
                              action='store_true',
                              help='list available example directories and exit')
    create_group.add_argument('-e', '--example',
                              type=str,
                              metavar='PATH',
                              help='example directory to copy')

    parser.add_argument('-d', '--destination',
                        type=str,
                        metavar='PATH',
                        default=os.getcwd(),
                        help="path to new project directory (default: %(default)s)")


def create(args):
    """
    Create a new ActivitySim configuration from an existing template.

    Use the -l flag to view a list of example configurations, then
    copy to your own working directory. These new project files can
    be run with the 'run' command.
    """

    if args.list:

        list_examples()
        sys.exit(0)

    if args.example:

        get_example(args.example, args.destination)
        sys.exit(0)


def list_examples():
    print('*** Available examples ***\n')

    ret = []
    for example in list(EXAMPLES.values()):
        del example['include']
        ret.append(example)
        print(yaml.dump(example))

    return ret


def get_example(example_name, destination):
    """
    Copy project data to user-specified directory.

    Examples and their data are described in a manifest
    YAML file. Each example contains at least a `name` and
    `include` field which is a list of files/folders to include
    in the copied example.



    Parameters
    ----------

    example_name: str, name of the example to copy.
        Options can be found via list_examples()
    destination: name of target directory to copy files to.
        If the target directory already exists, project files
        will be copied into a subdirectory with the same name
        as the example
    """
    if example_name not in EXAMPLES:
        sys.exit(f"error: could not find example '{example_name}'")

    if os.path.isdir(destination):
        dest_path = os.path.join(destination, example_name)
    else:
        dest_path = destination

    example = EXAMPLES[example_name]
    for item in example.get('include', []):

        # split include string into source/destination paths
        items = item.split()
        assets = items[0]
        if len(items) == 2:
            target_path = os.path.join(dest_path, items[-1])
        else:
            target_path = dest_path

        if assets.startswith('http'):
            download_asset(assets, target_path)

        else:
            for asset_path in glob.glob(_example_path(assets)):
                copy_asset(asset_path, target_path)

    print(f'copied! new project files are in {os.path.abspath(dest_path)}')
    print(f'the copied example can be run with\n')
    print(f'  activitysim run -w {dest_path}')


def copy_asset(asset_path, target_path):

    print(f'copying {os.path.basename(asset_path)} ...')
    if os.path.isdir(asset_path):
        target_path = os.path.join(target_path, os.path.basename(asset_path))
        shutil.copytree(asset_path, target_path)

    else:
        shutil.copy(asset_path, target_path)


def download_asset(url, target_path):

    print(f'downloading {os.path.basename(target_path)} ...')
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(target_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)
